# Import logger
from osaft.core.logger import log

# Import API
from osaft.core.backgroundfields import BackgroundField, WaveTypes
from osaft.core.fluids import InviscidFluid, ViscoelasticFluid, ViscousFluid
from osaft.core.solids import ElasticSolid, RigidSolid
from osaft.plotting.arf.arf_plots import ARFPlot
from osaft.plotting.scattering.fluid_plots import FluidScatteringPlot
from osaft.plotting.scattering.particle_plots import (
    ParticleScatteringPlot,
    ParticleWireframePlot,
)
# IMPORT API solutions
import osaft.solutions.Gorkov1962 as Gorkov1962
import osaft.solutions.King1934 as King1934
import osaft.solutions.Yosioka1955 as Yosioka1955
import osaft.solutions.Doinikov1994Rigid as Doinikov1994Rigid
import osaft.solutions.Doinikov1994Compressible as Doinikov1994Compressible
import osaft.solutions.Settnes2012 as Settnes2012



from collections.abc import Sequence
from enum import Enum
from numbers import Number
from typing import Optional, Union

import numpy as np

from osaft.core.functions import pi


class StringFormatter:
    """ Consistent Formatting for __str__ and __repr__ methods
    """

    @staticmethod
    def get_str_text(
        description: str, variable: str,
        value: Union[float, int, str], unit: Optional[str] = None,
        linebreak: bool = True,
    ) -> str:
        """
        Returns a constant formatting for the `__str__` method of the ARF
        classes

        :param description: Physical name for the variable
        :param variable: symbol/name of the variable
        :param value: value of the variable
        :param unit: unit of the variable
        :param linebreak: Appends a linebreak to the end of the text
        """
        out = f'{description:<20s}:\t'
        out += f'{variable:>8s} = '
        if isinstance(value, bool) or value is None:
            out += str(value)
        elif isinstance(value, Enum):
            out += f'{value.name:8s}'
        else:
            out += f'{value:5.2e}'
        if unit is not None and unit != '':
            out += f' [{unit:s}]'
        if linebreak:
            out += '\n'

        return out


class BaseValueTester:
    """ BaseClass for input value testing
    """

    low = None
    high = None
    msg = None

    @classmethod
    def test_arg(cls, arg: np.ndarray) -> None:
        """ Tests if (any value of) :attr:`arg` is between
        :attr:`cls.low` and :attr:`cls.high`

        :param arg: array to be tested
        :raises: ValueError if test fails
        """
        if np.any(arg > cls.high) > 0 or np.any(arg < cls.low):
            raise ValueError(cls.msg)


class ThetaTester(BaseValueTester):
    """ Test if  :math:`0 <= \\theta <= \\pi`
    """

    @classmethod
    def test_theta(cls, theta: Union[Number, np.ndarray]) -> None:
        """ Tests :attr:`theta` value is between 0 and :math:`\\pi`

        :param theta: input angles
        :raises: ValueError if  at least one theta is > pi or < 0
        """
        cls.low = 0
        cls.high = pi
        cls.msg = 'At least on theta values is greater '
        cls.msg += 'than pi or smaller than zero.'
        ThetaTester.test_arg(theta)


class RadiusTester(BaseValueTester):
    """ Test if  :math:`r` is inside or outside of sphere
    """

    @classmethod
    def is_inside_sphere(cls, r: Union[Number, np.array], R_0: float) -> None:
        """ Tests :attr:`r` value is between 0 and :attr:`R_0`

        :param r: input radii
        :param R_0: particle radius
        :raises: ValueError if  at least one theta is > R_0 or < 0
        """
        cls.low = 0
        cls.high = R_0
        cls.msg = 'At least one value of r is negative and/or greater '
        cls.msg += 'greater than R_0'
        RadiusTester.test_arg(r)

    @classmethod
    def is_outside_sphere(
        cls, r: Union[Number, np.ndarray],
        R_0: float,
    ) -> None:

        """ Tests :attr:`r` value is greater than :attr:`R_0`

        :param r: input radii
        :param R_0: particle radius
        :raises: ValueError if  at least one r < R_0
        """
        cls.low = R_0
        cls.high = np.Infinity
        cls.msg = 'At least one value of r is smaller than R_0'
        RadiusTester.test_arg(r)


class InputHandler:
    """ Handles inputs for velocity field methods

    Tests if inputs r, theta, t passed to methods that compute the velocity
    field are valid inputs and converts the inputs to numpy.ndarray.
    """

    test_theta = ThetaTester.test_theta
    test_r_inside = RadiusTester.is_inside_sphere
    test_r_outside = RadiusTester.is_outside_sphere

    @classmethod
    def handle_input(
        cls, r: Union[Number, Sequence],
        theta: Union[Number, Sequence],
        t: Union[Number, Sequence],
        R_0: Union[Number, Sequence],
        inside_sphere: bool,
    ):
        """Tests if inputs r, theta, t passed to methods that compute the
        velocity field are valid inputs and converts the
        inputs to numpy.ndarray.

        :param r: radial coordinate
        :param theta: tangential coordinate
        :param t: time
        :param R_0: radius of the sphere
        :param inside_sphere: if `True`/`False` tests if `r` is inside/outside
        """
        r = cls.handle_r(r, R_0, inside_sphere)
        theta = cls.handle_theta(theta)
        t = cls._to_array(t)

        return r, theta, t

    @classmethod
    def handle_r(cls, r, R_0, inside_sphere):
        """Tests if inputs `r` is inside or outside sphere and converts input
        to numpy.ndarray

        :param r: radial coordinate
        :param R_0: radius of the sphere
        :param inside_sphere: if `True`/`False` tests if `r` is inside/outside
        """
        r = cls._to_array(r)
        if inside_sphere:
            cls.test_r_inside(r, R_0)
        else:
            cls.test_r_outside(r, R_0)
        return r

    @classmethod
    def handle_theta(cls, theta):
        """Tests if inputs `theta` is between 0 and :math:`\\pi` input to
        numpy.ndarray

        :param theta: tangential coordinate
        """

        theta = cls._to_array(theta)
        cls.test_theta(theta)
        return theta

    @staticmethod
    def _to_array(arg: [Number, Sequence]):
        return np.asarray(arg)


if __name__ == '__main__':
    pass



import inspect
import unittest

from matplotlib import pyplot as plt

from osaft import ParticleScatteringPlot, ParticleWireframePlot
from osaft.tests.plotting.setup_test_scattering import BaseTestScattering


class TestParticleScattering(BaseTestScattering):

    def setUp(self) -> None:
        super().setUp()

        self.particle_plot = ParticleScatteringPlot(
            self.yosioka,
        )

        self.wireframe = ParticleWireframePlot(
            self.yosioka,
            scale_factor=0.1,
        )

    def test_tricontourf(self):
        fig, ax = self.particle_plot.plot(
            inst=True,
            phase=1,
            displacement=True,
            symmetric=True,
        )
        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_animation_tripcolor(self):
        anim = self.particle_plot.animate(
            tripcolor=True,
            frames=10,
            interval=100,
            plot_disp=False,
            mode=None,
        )
        anim.resume()
        plt.show(block=False)
        plt.pause(1)
        plt.close()

    def test_animation_tricontourf(self):
        anim = self.particle_plot.animate(
            frames=10,
            interval=100,
            plot_disp=True,
            mode=4,
        )
        anim.resume()
        plt.show(block=False)
        plt.pause(1)
        plt.close()

    def test_tripcolor(self):
        fig, ax = self.particle_plot.plot(
            tripcolor=True,
            inst=False,
            phase=1,
            plot_disp=False,
            symmetric=True,
        )
        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_tricontourf_evolution(self):
        fig, ax = self.particle_plot.plot_evolution(
            symmetric=True,
            displacement=True,
            mode=None,
        )
        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_evolution_tripcolor(self):
        fig, ax = self.particle_plot.plot_evolution(
            symmetric=True,
            mode=1,
            tripcolor=True,
        )
        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_wireframe_plot(self):
        fig, ax = self.wireframe.plot(phase=2)
        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_wireframe_evolution(self):
        fig, ax, = self.wireframe.plot_evolution()
        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_wireframe_animation(self):
        anim = self.wireframe.animate(mode=None, frames=10)
        anim.resume()
        plt.show(block=False)
        plt.pause(1)
        plt.close()


if __name__ == '__main__':
    unittest.main()



import inspect
import unittest

from matplotlib import pyplot as plt

from osaft import FluidScatteringPlot
from osaft.tests.plotting.setup_test_scattering import BaseTestScattering


class TestFluidScattering(BaseTestScattering):

    def setUp(self) -> None:

        super().setUp()

        r_max = 5 * self.R_0
        self.fluid_plot = FluidScatteringPlot(
            self.yosioka, r_max=r_max,
        )

    def test_tripcolor_symmetric(self):
        fig, ax = self.fluid_plot.plot(
            tripcolor=True,
            inst=False,
            mode=None,
            phase=0,
            symmetric=True,
            incident=True,
            scattered=True,
        )
        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_tripcolor_not_symmetric(self):
        fig, ax = self.fluid_plot.plot(
            tripcolor=True,
            inst=False,
            mode=None,
            phase=0,
            symmetric=False,
            incident=True,
            scattered=True,
        )

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_tricontourf(self):
        fig, ax = self.fluid_plot.plot(
            inst=True,
            mode=1,
            phase=0,
            symmetric=True,
            incident=False,
            scattered=True,
        )
        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_animation_tricontourf(self):
        anim = self.fluid_plot.animate(
            frames=10,
            interval=100,
            symmetric=True,
            scattered=False,
            incident=True,
            mode=None,
        )
        anim.resume()
        plt.show(block=False)
        plt.pause(1)
        plt.close()

    def test_animation_tripcolor(self):
        anim = self.fluid_plot.animate(
            tripcolor=True,
            frames=10,
            interval=100,
            symmetric=True,
            scattered=True,
            incident=True,
            mode=None,
        )
        anim.resume()
        plt.show(block=False)
        plt.pause(1)
        plt.close()

    def test_animation_tricontourf_not_symmetric(self):
        anim = self.fluid_plot.animate(
            frames=10,
            interval=100,
            symmetric=False,
            scattered=False,
            incident=True,
            mode=None,
        )
        anim.resume()
        plt.show(block=False)
        plt.pause(1)
        plt.close()

    def test_animation_tripcolor_not_symmetric(self):
        anim = self.fluid_plot.animate(
            tripcolor=True,
            frames=10,
            interval=100,
            symmetric=False,
            scattered=True,
            incident=True,
            mode=None,
        )
        anim.resume()
        plt.show(block=False)
        plt.pause(1)
        plt.close()

    def test_evolution_tricontourf(self):
        fig, ax = self.fluid_plot.plot_evolution(
            symmetric=True,
            scattered=True,
            incident=False,
            mode=1,
        )
        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_evolution_tripcolor(self):
        fig, ax = self.fluid_plot.plot_evolution(
            symmetric=True,
            scattered=True,
            incident=False,
            mode=None,
            tripcolor=True,
        )
        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Union

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation

from osaft.core.functions import pi, sqrt
from osaft.plotting.datacontainers.scattering_datacontainer import (
    FluidScatteringData,
)
from osaft.plotting.scattering.tri_plotter import TriangulationPlotter
from osaft.solutions.base_scattering import BaseScattering


class FluidScatteringPlot:
    """Class for plotting scattering field of the fluid

    Plots the acoustic field inside the particle using Matplotlib
    tricontourf or tripcolor plotting methods.

    :param sol: solution to be plotted
    :param r_max: radial limit of plot range
    :param theta_min: lower limit for tangential plot range
    :param theta_max: upper limit for tangential plot range
    :param resolution: if tuple (radial resolution, tangential resolution)
    :param cmap: color map
    """
    def __init__(
            self,
            sol: BaseScattering,
            r_max: float,
            theta_min: float = 0,
            theta_max: float = pi,
            resolution: Union[int, tuple[int, int]] = 100,
            cmap: str = 'winter',
    ):
        """Constructor method
        """
        self.data = FluidScatteringData(
            sol, sqrt(2) * r_max, theta_min,
            theta_max,
            res=resolution,
        )
        self.plotter = TriangulationPlotter(False, cmap)

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def plot(
            self,
            inst: bool = True,
            phase: float = 0,
            mode: Optional[int] = None,
            scattered: bool = True,
            incident: bool = True,
            symmetric: bool = True,
            tripcolor: bool = False,
            ax: Optional[plt.Axes] = None,
            **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Tricontourf plot for acoustic velocity field of the fluid

        Plots the velocity amplitude of the first-order acoustic velocity field
        of the fluid using Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tripcolor.html>`_
        if `tripcolor = True`

        :param inst: if `True` instantaneous amplitude is plotted
        :param phase: phase [0, 2 * pi]
        :param mode: mode of oscillation
        :param scattered: if `True` scattering field is plotted
        :param incident: if `True` incident field is plotted
        :param symmetric: if `True` the symmetry of the solution is used
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param ax: Axes object
        :param kwargs: passed through to tricontourf()
        """
        return self._triangulation_plot(
            tripcolor=tripcolor,
            inst=inst,
            phase=phase,
            mode=mode,
            scattered=scattered,
            incident=incident,
            symmetric=symmetric,
            ax=ax,
            **kwargs
        )

    def animate(
            self,
            frames: int = 64,
            interval: float = 100.0,
            mode: Optional[int] = None,
            scattered: bool = True,
            incident: bool = True,
            symmetric: bool = True,
            tripcolor: bool = False,
            ax: Optional[plt.Axes] = None,
            **kwargs,
    ) -> FuncAnimation:
        """Tricontourf animation for acoustic velocity field of the fluid

        Animates the velocity amplitude of the first-order acoustic velocity
        field of the fluid over one period using Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tripcolor.html>`_
        if `tripcolor = True`

        :param frames: number of frames for the animation
        :param interval: interval between frames in ms
        :param mode: mode of oscillation
        :param scattered: if `True` scattering field is plotted
        :param incident: if `True` incident field is plotted
        :param symmetric: if `True` the symmetry of the solution is used
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param ax: Axes object
        :param kwargs: passed through to tricontourf()
        """
        return self._triangulation_animation(
            tripcolor=tripcolor,
            frames=frames,
            interval=interval,
            mode=mode,
            scattered=scattered,
            incident=incident,
            symmetric=symmetric,
            ax=ax,
            **kwargs,
        )

    def plot_evolution(
            self,
            inst: bool = True,
            mode: Optional[int] = None,
            scattered: bool = True,
            incident: bool = True,
            symmetric: bool = True,
            tripcolor: bool = False,
            layout: tuple[int, int] = (3, 3),
            **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Tricontourf for acoustic velocity field evolution of the fluid

        Plots the velocity amplitude of the first-order acoustic velocity
        field of the fluid over one period at different phases using
        Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tripcolor.html>`_
        if `tripcolor = True`.

        The first phase value is always :math:`0\\pi` and the last one
        :math:`2\\pi`. The total number of plots and, hence, also the steps
        between the different phase values is the defined by the product of the
        ``layout`` tuple.

        :param inst: if `True` instantaneous amplitude is plotted
        :param mode: mode of oscillation
        :param scattered: if `True` scattering field is plotted
        :param incident: if `True` incident field is plotted
        :param symmetric: if `True` the symmetry of the solution is used
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param layout: number of rows and columns for plotting
        :param kwargs: passed through to the parent subplots command
        """

        n_row, n_col = layout
        n = n_col * n_row

        phases = np.linspace(0, 2, num=n)

        fig, axes = plt.subplots(
            n_row, n_col,
            sharex=True, sharey=True,
            **kwargs,
        )

        # Get velocity norm, needed to make colormap of the right range
        X, Y, C_norm = self.data.get_velocity_magnitude(
            False, mode=mode, scattered=scattered, incident=incident,
        )
        # Only values inside the plotting range
        C_norm = np.where(
            np.hypot(X, Y) < self.data.r_max,
            C_norm, 0,
        )

        for i, phase in enumerate(phases):

            row = i // n_col
            col = i % n_col
            ax = axes.flat[i]

            X, Z, C = self.data.get_velocity_magnitude(
                instantaneous=inst,
                phase=phase * np.pi,
                mode=mode,
                scattered=scattered,
                incident=incident,
            )
            # Color bar label
            cbar_label = 'Acoustic Velocity [m/s]'

            # Plot
            _, _, cnf, cbar, _ = self.plotter.plot(
                X=Z,
                Y=X,
                C=C,
                radius=self.data.sol.R_0,
                symmetric=symmetric,
                tripcolor=tripcolor,
                cbar_label=cbar_label,
                ax=ax,
                vmin=0,
                vmax=1.01 * C_norm.max(),
            )
            # remove colorbar
            cbar.remove()

            ticks = self.data.sol.R_0 * np.asarray([-1, 1])
            labels = [-1, 1]

            ax.set_title(f'{phase:.2f}' + r'$\pi$')

            if row != (n_row - 1):
                ax.set_xlabel('')
            if col > 0:
                ax.set_ylabel('')

            ax.set_xticks(ticks, labels=labels)
            ax.set_yticks(ticks, labels=labels)

            # set aspect ratio to 1:1
            ax.set_aspect(1)

        fig.tight_layout()
        cbar = fig.colorbar(cnf, ax=axes.ravel().tolist())
        cbar_label = 'Acoustic Velocity [m/s]'
        cbar.ax.set_ylabel(cbar_label)

        return fig, axes

    # -------------------------------------------------------------------------
    # Private Methods
    # -------------------------------------------------------------------------

    def _triangulation_plot(
            self,
            tripcolor: bool,
            inst: bool,
            phase: float,
            mode: Optional[int],
            scattered: bool,
            incident: bool,
            symmetric: bool,
            ax: Optional[plt.Axes],
            **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Helper function for tripcolor/tricontourf plot

        :param tripcolor: if `True` tripcolor, else tricontourf plot
        :param inst: if `True` instantaneous amplitude is plotted
        :param phase: phase [0, 2 * pi]
        :param mode: mode of oscillation
        :param scattered: if `True` scattering field is plotted
        :param incident: if `True` incident field is plotted
        :param symmetric: if `True` the symmetry of the solution is used
        :param ax: Axes object
        :param kwargs: passed through to plotting method
        """

        # Data
        X, Z, C = self.data.get_velocity_magnitude(
            instantaneous=inst,
            phase=phase,
            mode=mode,
            scattered=scattered,
            incident=incident,
        )
        # Color bar label
        cbar_label = 'Acoustic Velocity [m/s]'

        # Plot
        fig, ax, _, _, _ = self.plotter.plot(
            X=Z,
            Y=X,
            C=C,
            radius=self.data.sol.R_0,
            symmetric=symmetric,
            tripcolor=tripcolor,
            cbar_label=cbar_label,
            ax=ax,
            **kwargs,
        )
        return fig, ax

    def _triangulation_animation(
            self,
            tripcolor: bool,
            frames: int,
            interval: float,
            mode: Optional[int] = None,
            scattered: bool = True,
            incident: bool = True,
            symmetric: bool = True,
            ax: Optional[plt.Axes] = None,
            **kwargs,
    ) -> FuncAnimation:
        """Helper function for tripcolor/tricontourf animation

        :param tripcolor: if `True` tripcolor, else tricontourf plot
        :param frames: number of frames for the animation
        :param interval: interval between frames in ms
        :param mode: mode of oscillation
        :param scattered: if `True` scattering field is plotted
        :param incident: if `True` incident field is plotted
        :param symmetric: if `True` the symmetry of the solution is used
        :param ax: Axes object
        :param kwargs: passed through to tricontourf()
        """
        # Data function for animation
        def data_func(phase: float) -> tuple[
            np.ndarray, np.ndarray,
            np.ndarray,
        ]:
            """Returns the velocity field for given phase
            Closure is used to fix all additional parameters
            :param phase: phase
            """
            return self.data.get_velocity_magnitude(
                True, phase, mode,
                scattered, incident,
            )
        # Color bar label
        cbar_label = 'Acoustic Velocity [m/s]'

        # Get velocity norm, needed to make colormap of the right range
        X, Y, C_norm = self.data.get_velocity_magnitude(
            False, mode=mode, scattered=scattered, incident=incident,
        )
        # Only values inside the plotting range
        C_norm = np.where(
            np.hypot(X, Y) < self.data.r_max,
            C_norm, 0,
        )

        return self.plotter.animate(
            frames=frames,
            interval=interval,
            tripcolor=tripcolor,
            symmetric=symmetric,
            cbar_label=cbar_label,
            data_func=data_func,
            C_norm=C_norm,
            radius=self.data.sol.R_0,
            ax=ax,
            **kwargs
        )


if __name__ == '__main__':
    pass



import unittest

from osaft.core.fluids import InviscidFluid, ViscoelasticFluid, ViscousFluid
from osaft.core.frequency import Frequency
from osaft.core.functions import pi, sqrt
from osaft.tests.basetest import BaseTest


class TestInviscidFluid(BaseTest):

    def setUp(self) -> None:

        super().setUp()

        self.cls = InviscidFluid(
            self.f,
            self.rho_f,
            self.c_f,
        )
        self.composite_cls = InviscidFluid(
            self.frequency,
            self.rho_f, self.c_f,
        )

        self.list_cls = [self.cls, self.composite_cls]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def compute_k_f(self) -> float:
        omega = self.frequency.omega
        c_f = self.c_f
        return omega / c_f

    def compute_kappa_f(self) -> float:
        return 1 / self.c_f**2 / self.rho_f

    def compute_lambda_f(self) -> float:
        return self.c_f / self.f

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_omega(self) -> None:
        self.do_testing(
            func_1=lambda: self.frequency.omega,
            func_2=lambda: self.cls.omega,
        )
        self.do_testing(
            func_1=lambda: self.frequency.omega,
            func_2=lambda: self.composite_cls.omega,
        )

    def test_k_f(self) -> None:
        self.do_testing(
            func_1=self.compute_k_f,
            func_2=lambda: self.cls.k_f,
        )
        self.do_testing(
            func_1=self.compute_k_f,
            func_2=lambda: self.composite_cls.k_f,
        )

    def test_lambda_f(self) -> None:
        self.do_testing(
            func_1=self.compute_lambda_f,
            func_2=lambda: self.cls.lambda_f,
        )
        self.do_testing(
            func_1=self.compute_lambda_f,
            func_2=lambda: self.composite_cls.lambda_f,
        )

    def test_kappa_f(self) -> None:
        self.do_testing(
            func_1=self.compute_kappa_f,
            func_2=lambda: self.cls.kappa_f,
        )
        self.do_testing(
            func_1=self.compute_kappa_f,
            func_2=lambda: self.composite_cls.kappa_f,
        )


class TestViscousFluid(TestInviscidFluid):

    def setUp(self) -> None:

        BaseTest.setUp(self)

        self.cls = ViscousFluid(
            self.f, self.rho_f, self.c_f,
            self.eta_f, self.zeta_f,
        )
        self.composite_cls = ViscousFluid(
            self.frequency,
            self.rho_f, self.c_f,
            self.eta_f, self.zeta_f,
        )
        self.cls_inviscid = InviscidFluid(
            self.f, self.rho_f,
            self.c_f,
        )

        self.list_cls = [self.cls, self.composite_cls, self.cls_inviscid]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def compute_k_f(self) -> complex:

        omega = self.cls.omega
        c = self.c_f
        rho = self.rho_f
        eta = self.eta_f
        zeta = self.zeta_f
        return omega / c / sqrt(
            1 - 1j * omega / (rho * c**2)
            * (zeta + 4 * eta / 3),
        )

    def compute_k_v(self) -> complex:
        omega = self.cls.omega
        eta = self.eta_f
        rho = self.rho_f
        return (1 + 1j) * sqrt(rho * omega / (2 * eta))

    def compute_delta(self) -> complex:
        return 1 / self.compute_k_v().imag

    def compute_lambda_v(self) -> complex:
        return 2 * pi / self.compute_k_v().real

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_delta(self) -> None:
        self.do_testing(
            func_1=self.compute_delta,
            func_2=lambda: self.cls.delta,
        )
        self.do_testing(
            func_1=self.compute_delta,
            func_2=lambda: self.composite_cls.delta,
        )

    def test_delta_inviscid(self) -> None:
        self.cls.eta_f = 0
        self.assertEqual(self.cls.delta, 0)

    def test_lambda_v(self) -> None:
        self.do_testing(
            func_1=self.compute_lambda_v,
            func_2=lambda: self.cls.lambda_v,
        )
        self.do_testing(
            func_1=self.compute_lambda_v,
            func_2=lambda: self.composite_cls.lambda_v,
        )

    def test_lambda_v_inviscid(self) -> None:
        self.cls.eta_f = 0
        self.assertEqual(self.cls.lambda_v, 0)

    def test_k_v(self) -> None:
        self.do_testing(
            func_1=self.compute_k_v,
            func_2=lambda: self.cls.k_v,
        )
        self.do_testing(
            func_1=self.compute_k_v,
            func_2=lambda: self.composite_cls.k_v,
        )

    def test_k_v_inviscid(self) -> None:
        self.cls.eta_f = 0
        self.assertEqual(self.cls.k_v, 0)

    def test_k_f_inviscid(self):
        self.cls.eta_f = 0
        self.cls.zeta_f = 0
        self.assertAlmostEqual(self.cls.k_f, self.cls_inviscid.k_f)


class TestViscoelasticFluid(TestViscousFluid):

    def setUp(self) -> None:

        BaseTest.setUp(self)

        self.frequency = Frequency(self.f)
        self.cls = ViscoelasticFluid(
            self.f, self.rho_f,
            self.c_f,
            self.eta_f, self.eta_p,
            self.zeta_f, self.zeta_p,
            self.lambda_M,
        )

        self.composite_cls = ViscoelasticFluid(
            self.frequency,
            self.rho_f,
            self.c_f,
            self.eta_f,
            self.eta_p,
            self.zeta_f,
            self.zeta_p,
            self.lambda_M,
        )

        self.cls_inviscid = InviscidFluid(
            self.f, self.rho_f,
            self.c_f,
        )

        self.list_cls = [self.cls, self.composite_cls, self.cls_inviscid]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def compute_eta_c(self):
        omega = self.cls.omega
        eta_f = self.eta_f
        eta_p = self.eta_p
        lambda_M = self.lambda_M
        return eta_f + eta_p / (1 - 1j * omega * lambda_M)

    def compute_zeta_c(self):
        omega = self.cls.omega
        zeta_f = self.zeta_f
        zeta_p = self.zeta_p
        lambda_M = self.lambda_M
        return zeta_f + zeta_p / (1 - 1j * omega * lambda_M)

    def compute_k_f(self) -> complex:
        omega = self.cls.omega
        c = self.c_f
        rho = self.rho_f
        eta = self.compute_eta_c()
        zeta = self.compute_zeta_c()
        return omega / c / sqrt(
            1 - 1j * omega / (rho * c**2)
            * (zeta + 4 * eta / 3),
        )

    def compute_k_v(self) -> complex:
        omega = self.cls.omega
        eta = self.compute_eta_c()
        rho = self.rho_f
        out = (1 + 1j) * sqrt(rho * omega / (2 * eta))
        return out

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_eta_c(self):
        self.do_testing(
            func_1=self.compute_eta_c,
            func_2=lambda: self.cls.eta_c,
        )
        self.do_testing(
            func_1=self.compute_eta_c,
            func_2=lambda: self.composite_cls.eta_c,
        )

    def test_zeta_c(self):
        self.do_testing(
            func_1=self.compute_zeta_c,
            func_2=lambda: self.cls.zeta_c,
        )
        self.do_testing(
            func_1=self.compute_zeta_c,
            func_2=lambda: self.composite_cls.zeta_c,
        )

    def test_k_v_inviscid(self) -> None:
        self.cls.eta_f = 0
        self.cls.eta_p = 0
        self.assertEqual(self.cls.k_v, 0)

    def test_delta_inviscid(self) -> None:
        self.cls.eta_f = 0
        self.cls.eta_p = 0
        self.assertEqual(self.cls.delta, 0)

    def test_lambda_v_inviscid(self) -> None:
        self.cls.eta_f = 0
        self.cls.eta_p = 0
        self.assertEqual(self.cls.lambda_v, 0)

    def test_k_f_inviscid(self):
        self.cls.eta_f = 0
        self.cls.zeta_f = 0
        self.cls.eta_p = 0
        self.cls.zeta_p = 0
        self.assertAlmostEqual(self.cls.k_f, self.cls_inviscid.k_f)


if __name__ == '__main__':
    unittest.main()



from collections.abc import Sequence
from typing import Optional, Union

from osaft import log
from osaft.core.backgroundfields import WaveTypes
from osaft.core.frequency import Frequency
from osaft.core.functions import SpecialFunctions as Sp
from osaft.core.functions import full_range
from osaft.core.geometries import Sphere
from osaft.core.helper import InputHandler
from osaft.core.helper import StringFormatter as SF
from osaft.core.variable import ActiveVariable
from osaft.solutions.base_scattering import BaseScattering
from osaft.solutions.Yosioka1955.base import BaseYosioka


class ScatteringField(BaseYosioka, BaseScattering):
    """Scattering field class for Yosioka & Kawasima (1955)

    :param frequency: Frequency [Hz]
    :param radius: Radius of the sphere [m]
    :param rho_s: Density of the fluid-like sphere [kg/m^3]
    :param c_s: Speed of sound of the particle [m/s]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param position: Position in the standing wave field [rad]
    :param wave_type: either standing or progressive wave
    """

    def __init__(
        self, frequency: Union[Frequency, float],
        radius: Union[Sphere, float],
        rho_s: float, c_s: float, rho_f: float, c_f: float,
        p_0: float, wave_type: WaveTypes,
        position: Optional[float] = None,
        N_max: int = 5,
    ) -> None:
        """Constructor method
        """

        BaseScattering.__init__(self, N_max)
        BaseYosioka.__init__(
            self, frequency, radius,
            rho_s, c_s, rho_f, c_f,
            p_0, wave_type, position,
        )

        log.info('Create ScatteringField')

        # Dependent Variables
        self._A_n = ActiveVariable(
            self._reset_coeff,
            'Coefficient A_n'
            ' for scattered potential',
        )

        self._B_n = ActiveVariable(
            self._reset_coeff,
            'Coefficient B_n'
            ' for potential inside particle',
        )

        # Dependencies
        self._A_n.is_computed_by(
            self._labmda_rho, self.fluid._k_f,
            self.scatterer._k_f, self.sphere._R_0, self.field._A_in,
        )
        self._B_n.is_computed_by(
            self._labmda_rho, self.fluid._k_f,
            self.scatterer._k_f, self.sphere._R_0, self.field._A_in,
        )

        if type(self) is ScatteringField:
            log.info(str(self))
            log.debug(repr(self))

    def __repr__(self):
        return (
            f'{type(self)}(frequency={self.f}, radius={self.R_0}, '
            f'rho_s={self.rho_s}, rho_f={self.rho_f}, '
            f'c_f={self.c_f}, c_s={self.c_s}'
            f'p_0={self.p_0}, position={self.position}, '
            f'wave_type={self.wave_type}, N_max={self.N_max})'
        )

    def __str__(self):
        out = 'Yosioka\'s selfution to the Scattering field'
        out += ' with following properties: \n'
        out += SF.get_str_text('Wave type', '', self.wave_type, '')
        out += SF.get_str_text('Frequency', 'f', self.f, 'Hz')
        out += SF.get_str_text('Pressure', 'p_0', self.p_0, 'Pa')
        out += SF.get_str_text(
            'Position', 'd',
            self.position, 'Pa',
        )
        out += SF.get_str_text(
            'Wavelength', 'lambda',
            self.field.lambda_f, 'm',
        )

        out += 'Fluid\n'
        out += SF.get_str_text(
            'Density', 'rho_f',
            self.rho_f, 'kg/m^3',
        )
        out += SF.get_str_text(
            'Sound Speed', 'c_f',
            self.c_f, 'm/s',
        )
        out += SF.get_str_text(
            'Compressibility', 'kappa_f',
            self.kappa_f, '1/Pa',
        )

        out += 'Particle\n'
        out += SF.get_str_text(
            'Radius', 'R_0',
            self.R_0, 'm',
        )
        out += SF.get_str_text(
            'Density', 'rho_s',
            self.rho_s, 'kg/m^3',
        )
        out += SF.get_str_text(
            'Speed of sound', 'c_s',
            self.c_s, 'm/s',
        )
        out += 'Other\n'
        out += SF.get_str_text(
            '#modes', 'N_max',
            self.N_max, None,
        )
        return out

    # -------------------------------------------------------------------------
    # User-faced functions
    # -------------------------------------------------------------------------

    def radial_particle_velocity(
        self,
        r: Union[float, Sequence[float]],
        theta: Union[float, Sequence[float]],
        t: Union[float, Sequence[float]],
        mode: Optional[int] = None,
    ) -> Union[float, Sequence[float]]:
        """Radial particle velocity [m/s]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode
        """

        def radial_func(l, x):
            return self.B_n(l) * self.k_s * Sp.d1_besselj(l, self.k_s * x)
        r, theta, t = InputHandler.handle_input(
            r, theta, t, self.R_0,
            inside_sphere=True,
        )
        out = self.radial_mode_superposition(radial_func, r, theta, t, mode)
        return out

    def tangential_particle_velocity(
        self,
        r: Union[float, Sequence[float]],
        theta: Union[float, Sequence[float]],
        t: Union[float, Sequence[float]],
        mode: Optional[int] = None,
    ) -> Union[float, Sequence[float]]:
        """Tangential particle velocity [m/s]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode
        """

        def radial_func(l, x):
            return self.B_n(l) * Sp.besselj(l, self.k_s * x) / x
        r, theta, t = InputHandler.handle_input(
            r, theta, t, self.R_0,
            inside_sphere=True,
        )
        out = self.tangential_mode_superposition(
            radial_func, r, theta, t,
            mode,
        )
        return out

    # -------------------------------------------------------------------------
    # Velocity Amplitudes and Velocity Potentials
    # -------------------------------------------------------------------------

    def Phi_1(
        self, r: float, theta: float, t: float,
        mode: Optional[int] = None,
    ) -> complex:
        r"""Fluid velocity potential :math:`\Phi_1` [m^2/s]

        (Eq. 17)

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode
        """

        return self.Phi_i(r, theta, t, mode) + self.Phi_s(r, theta, t, mode)

    def Phi_i(
        self, r: float, theta: float, t: float,
        mode: Optional[int] = None,
    ) -> complex:
        r"""Fluid velocity potential of the incident field :math:`\Phi_i`
        [m^2/s]

        (Eq. 16, 27)

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode
        """

        def radial_func(l, x):
            return self.field.A_in(l) * Sp.besselj(l, self.k_f * x)

        out = self.radial_mode_superposition(radial_func, r, theta, t, mode)
        return out

    def Phi_s(
        self, r: float, theta: float, t: float,
        mode: Optional[int] = None,
    ) -> complex:
        r"""Fluid velocity potential of the scattered field :math:`\Phi_s`
        [m^2/s]

        (Eq. 18, 29)

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode
        """
        def radial_func(l, x):
            return self.A_n(l) * Sp.hankelh2(l, self.k_f * x)

        out = self.radial_mode_superposition(radial_func, r, theta, t, mode)
        return out

    def Phi_star(
        self, r: float, theta: float, t: float,
        mode: Optional[int] = None,
    ) -> complex:
        r"""Particle velocity potential :math:`\Phi^*` [m^2/s]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode
        """
        def radial_func(l, x):
            return self.B_n(l) * Sp.besselj(l, self.k_s * x)

        out = self.radial_mode_superposition(radial_func, r, theta, t, mode)
        return out

    def V_r_sc(
        self, n: int,
        r: Union[float, Sequence[float]],
    ) -> Union[complex, Sequence[complex]]:
        """Radial scattering field velocity term of mode `n`
        without Legendre coefficients

        Returns radial scattering field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """
        return (
            self.A_n(n)
            * self.k_f * Sp.d1_hankelh2(n, self.k_f * r)
        )

    def V_theta_sc(
        self, n: int,
        r: Union[float, Sequence[float]],
    ) -> Union[complex, Sequence[complex]]:
        """Tangential scattering field velocity term of mode n
        without Legendre coefficients

        Returns tangential scattering field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """
        return self.A_n(n) * Sp.hankelh2(n, self.k_f * r) / r

    # -------------------------------------------------------------------------
    # Coefficients
    # -------------------------------------------------------------------------

    def A_n(self, n: int) -> complex:
        """Coefficient :math:`A_n` [m^2/s]

        (Eq. 22)

        :param n: mode [-]
        """

        if n < len(self._A_n.value):
            return self._A_n.value[n]
        else:
            self._compute_A_n(n)
            return self._A_n.value[n]

    def B_n(self, n: int) -> complex:
        """Coefficient :math:`B_n` [m^2/s]

        (Eq. 23)

        :param n: mode [-]
        """

        if n < len(self._B_n.value):
            return self._B_n.value[n]
        else:
            self._compute_B_n(n)
            return self._B_n.value[n]

    def _compute_A_n(self, N: int):
        """ Compute A_n according to (22), multiplied with the incident
        amplitude
        """
        n_old = len(self._A_n.value)
        for n in full_range(n_old, N):
            term1 = (
                self.lambda_rho * self.k_f *
                Sp.besselj(n, self.x_s) * Sp.d1_besselj(n, self.x_f)
            )
            term2 = (
                self.k_s * Sp.d1_besselj(n, self.x_s)
                * Sp.besselj(n, self.x_f)
            )
            term3 = (
                self.k_s * Sp.d1_besselj(n, self.x_s)
                * Sp.hankelh2(n, self.x_f)
            )
            term4 = (
                self.lambda_rho * self.k_f *
                Sp.besselj(n, self.x_s) * Sp.d1_hankelh2(n, self.x_f)
            )
            out = (term1 - term2) / (term3 - term4)
            out *= self.A_in(n)
            self._A_n.value.append(out)

    def _compute_B_n(self, N: int):
        """Compute B_n according to (23), multiplied with the incident
        amplitude
        """
        n_old = len(self._B_n.value)
        for n in full_range(n_old, N):
            out = 1j * self.k_f
            out /= self.x_f ** 2
            term1 = (
                self.k_s * Sp.d1_besselj(n, self.x_s)
                * Sp.hankelh2(n, self.x_f)
            )
            term2 = (
                self.lambda_rho * self.k_f *
                Sp.besselj(n, self.x_s) * Sp.d1_hankelh2(n, self.x_f)
            )
            out /= (term1 - term2)
            out *= self.A_in(n)
            self._B_n.value.append(out)


if __name__ == '__main__':
    pass



import unittest
import warnings

from osaft import Doinikov1994Rigid
from osaft.core.backgroundfields import WaveTypes, WrongWaveTypeError
from osaft.core.functions import pi, sin, xexpexp1
from osaft.tests.basetest_arf import HelperStandingARF, HelperTravelingARF
from osaft.tests.Doinikov1994Rigid.test_base import TestBase


class TestARFBase(TestBase):

    def setUp(self) -> None:

        super().setUp()

        self.small_particle_limit = True
        self.small_viscous_boundary_layer = False
        self.large_viscous_boundary_layer = False
        self.fastened_sphere = False
        self.N_max = 5

        self._eta_f.low = 1e-5
        self._eta_f.high = 1e-3
        self._f.low = 1e6
        self._f.high = 10e6
        self._R_0.low = 1e-6

        self.cls = Doinikov1994Rigid.ARF(
            self.f,
            self.R_0, self.rho_s,
            self.rho_f, self.c_f, self.eta_f, self.zeta_f,
            self.p_0, WaveTypes.STANDING,
            self.position,
            self.small_particle_limit,
            self.small_viscous_boundary_layer,
            self.large_viscous_boundary_layer,
            self.fastened_sphere,
            self.N_max,
        )

        self.list_cls = [self.cls]

    def assign_parameters(self) -> None:
        super().assign_parameters()

        self.cls.fastened_sphere = self.fastened_sphere
        self.cls.small_viscous_boundary_layer = (
            self.small_viscous_boundary_layer
        )
        self.cls.large_viscous_boundary_layer = (
            self.large_viscous_boundary_layer
        )
        self.cls.small_particle_limit = self.small_particle_limit
        self.cls.N_max = self.N_max

    # ------------------

    def test_fastened_sphere(self) -> None:
        self.assign_parameters()
        self.assertEqual(self.fastened_sphere, self.cls.fastened_sphere)
        self.fastened_sphere = True
        self.assign_parameters()
        self.assertEqual(self.fastened_sphere, self.cls.fastened_sphere)

    def test_NotImplemented_solution(self) -> None:
        self.cls.wave_type = 'random'

        self.assertRaises(
            WrongWaveTypeError,
            self.cls.acoustic_radiation_force,
        )
    # ------------------

    def test_small_particle_limit(self) -> None:
        self.assign_parameters()
        self.assertEqual(
            self.small_particle_limit,
            self.cls.small_particle_limit,
        )

    def test_small_particle_limit_error(self) -> None:
        self.small_particle_limit = False
        self.assertRaises(NotImplementedError, self.assign_parameters)

    def test_small_viscous_boundary_layer(self) -> None:
        self.assign_parameters()
        self.assertEqual(
            self.small_viscous_boundary_layer,
            self.cls.small_viscous_boundary_layer,
        )
        self.small_viscous_boundary_layer = True
        self.assign_parameters()
        self.assertEqual(
            self.small_viscous_boundary_layer,
            self.cls.small_viscous_boundary_layer,
        )

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------

    @property
    def rho_t(self) -> float:
        return self.rho_f / self.rho_s

    @property
    def x_0(self) -> complex:
        return self.cls.field.k_f.real * self.R_0

    @property
    def x(self) -> complex:
        return self.cls.x

    @property
    def x_v(self) -> complex:
        return self.cls.x_v

    @property
    def norm_delta(self) -> complex:
        return self.cls.delta / self.R_0

    @property
    def physical_position(self) -> float:
        return self.position * self.c_f / self.cls.omega
    # -------------------------------------------------------------------------
    # F2 (Streaming contribution)
    # -------------------------------------------------------------------------

    @property
    def F_2_standing(self):
        k = self.cls.k_f
        x = self.x
        x_v = self.x_v
        d = self.cls.abs_pos
        first = sin(k * d + k.conjugate() * d) * sin(x + x.conjugate())
        first *= (x - x.conjugate()) / (x + x.conjugate())
        second = sin(k * d - k.conjugate() * d) * sin(x - x.conjugate())
        second *= (x + x.conjugate()) / (x - x.conjugate())
        out = first - second
        out *= - 3 / 4 * pi * self.rho_f * self.cls.field.abs_A_squared
        out *= x * x.conjugate()
        out /= x_v ** 2
        return out

    @property
    def F_2_travelling(self):
        x = self.x
        x_v = self.x_v

        out = sin(x - x.conjugate()) / (x - x.conjugate()) * 1j / x_v ** 2
        out *= x * x.conjugate() * (x + x.conjugate())
        out *= - 3 / 2 * pi * self.rho_f * self.cls.field.abs_A_squared
        return out

    # -------------------------------------------------------------------------
    # D_n
    # -------------------------------------------------------------------------
    @property
    def D_0(self):
        out = self.G_1
        out += self.x_v ** 3 * (12 + self.x_v ** 2) * self.function(self.x_v)
        out *= -self.x ** 3 * self.G_0
        out += 2 * self.x ** 3 / 9
        out += self.x ** 3 / (3 * self.x_v ** 2)
        return out

    @property
    def D_1(self):
        out = - 2 * (1 + 1j) * (9 + self.x_v ** 2)
        out *= self.function(self.x_v - 1j * self.x_v)
        out += 2 * self.G_3 * self.function(self.x_v)
        out += 1j * self.G_4 * self.function(-1j * self.x_v)
        out *= - self.x_v ** 3
        out += self.G_2
        out *= self.cls.x ** 3 * self.G_0.conjugate() / (1 + self.x_v)
        out += self.cls.x ** 3 / (3 * self.x_v ** 2)
        return out

    @staticmethod
    def function(z):
        return xexpexp1(z)

    # -------------------------------------------------------------------------
    # G_n
    # -------------------------------------------------------------------------

    @property
    def G_0(self) -> complex:
        denominator = 72 * (
            9 * self.rho_t + 9 * self.rho_t * self.x_v
            + (2 + self.rho_t) * self.x_v ** 2
        )
        numerator = (1 - self.rho_t)
        return numerator / denominator

    @property
    def G_1(self) -> complex:
        out = 48 - 96 * self.x_v + 2 * self.x_v ** 2 - 14 * self.x_v ** 3
        out += self.x_v ** 4 - self.x_v ** 5
        return out

    @property
    def G_2(self):
        out = 48 + (48 + 192 * 1j / 5) * self.x_v
        out += (122 + 192 * 1j) * self.x_v ** 2 / 5
        out += 42 * (1 - 1j) * self.x_v ** 3 / 5
        out += (49 + 36 * 1j) * self.x_v ** 4 / 10
        out -= (31 - 17 * 1j) * self.x_v ** 5 / 10
        out += (6 + 31 * 1j) * self.x_v ** 6 / 30
        out += (1 + 6 * 1j) * self.x_v ** 7 / 30
        out += 1j * self.x_v ** 8 / 30
        return out

    @property
    def G_3(self):
        return 3 - 3 * 1j * self.x_v - self.x_v ** 2

    @property
    def G_4(self):
        out = 9 + 9 * self.x_v + 41 * self.x_v ** 2 / 10
        out += 11 * self.x_v ** 3 / 10
        out += self.x_v ** 4 / 5 + self.x_v ** 5 / 30
        return out

    # -------------------------------------------------------------------------
    # Test Properties
    # -------------------------------------------------------------------------

    def test_D_1(self):
        warnings.filterwarnings('error', category=RuntimeWarning)
        try:
            self.do_testing(
                func_1=lambda: self.D_1,
                func_2=lambda: self.cls.D_1,
                threshold=1e-3,
            )
        except RuntimeWarning:
            pass
        warnings.filterwarnings('default', category=RuntimeWarning)

    def test_properties(self) -> None:
        properties = [
            'x_0', 'norm_delta', 'G_0', 'G_1', 'G_2', 'G_3', 'G_4',
            'D_0', 'F_2_standing', 'F_2_travelling',
        ]
        self._test_properties(properties, threshold=1e-6)


# |x| << 1  |x| << |x_v|

class TestSmallParticleTravelling(HelperTravelingARF, TestARFBase):

    def setUp(self):
        TestARFBase.setUp(self)
        self._threshold = 1e-7
        self.fastened_sphere = False
        self.large_viscous_boundary_layer = False
        self.small_viscous_boundary_layer = False
        self.assign_parameters()

    def compute_arf(self):
        F1 = self.D_0 - self.D_0.conjugate()
        F1 += 2 * (self.D_1 - self.D_1.conjugate())
        F1 *= 3 / 2 * 1j * pi * self.rho_f * self.cls.field.abs_A_squared
        F2 = self.F_2_travelling
        return (F1 + F2).real


class TestSmallParticleStanding(HelperStandingARF, TestARFBase):

    def setUp(self):
        TestARFBase.setUp(self)
        self._threshold = 1e-7
        self.fastened_sphere = False
        self.large_viscous_boundary_layer = False
        self.small_viscous_boundary_layer = False
        self.assign_parameters()

    def compute_arf(self):
        d = self.cls.abs_pos
        sine = sin(2 * self.cls.k_f * d)
        sine_conj = sin(2 * self.cls.k_f.conjugate() * d)
        F1 = self.D_0 * sine + self.D_0.conjugate() * sine_conj
        F1 -= 2 * (self.D_1 * sine + self.D_1.conjugate() * sine_conj)
        F1 *= 3 / 4 * pi * self.rho_f * self.cls.field.abs_A_squared
        F2 = self.F_2_standing
        return (F1 + F2).real


# |x| << 1 << |x_v|

class TestARFLimit1Standing(HelperStandingARF, TestARFBase):

    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.fastened_sphere = False
        self.large_viscous_boundary_layer = False
        self.small_viscous_boundary_layer = True
        self.assign_parameters()

    def compute_arf(self):
        out = (5 - 2 * self.rho_t) / (3 * (2 + self.rho_t))
        out += 3 * (1 - self.rho_t)**2 / (2 + self.rho_t)**2 * self.norm_delta

        out *= pi * self.rho_f * self.x_0**3 * self.cls.field.abs_A_squared
        out *= sin(2 * self.position)

        return out


class TestARFLimit1Traveling(HelperTravelingARF, TestARFBase):

    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.fastened_sphere = False
        self.large_viscous_boundary_layer = False
        self.small_viscous_boundary_layer = True
        self.assign_parameters()

    def compute_arf(self):
        out = 2 + 12 * self.rho_t - 21 * self.rho_t**2 + 7 * self.rho_t**3
        out *= -self.norm_delta / (2 + self.rho_t)

        out += (1 - self.rho_t)**2

        out *= 6 * pi * self.rho_f * self.x_0**3 * self.norm_delta
        out *= self.cls.field.abs_A_squared / (2 + self.rho_t)**2

        return out


# |x| << |x_v| << 1

class TestARFLimit2Standing(HelperStandingARF, TestARFBase):

    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.fastened_sphere = False
        self.large_viscous_boundary_layer = True
        self.small_viscous_boundary_layer = False
        self.assign_parameters()

    def compute_arf(self):
        out = 11 * (1 - self.rho_t)
        out /= 5 * self.norm_delta

        out += 2 * self.rho_t - 1

        out *= pi / 3 * self.rho_s * self.x_0**3 * self.cls.field.abs_A_squared
        out *= sin(2 * self.position)

        return out


class TestARFLimit2Traveling(HelperTravelingARF, TestARFBase):

    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.fastened_sphere = False
        self.large_viscous_boundary_layer = True
        self.small_viscous_boundary_layer = False
        self.assign_parameters()

    def compute_arf(self):
        out = -22 / 15 * pi * self.rho_s * self.x_0**3
        out *= self.cls.field.abs_A_squared

        out *= (1 - self.rho_t) / self.norm_delta

        return out


# |x| << 1 << |x_v|

class TestARFLimit3Standing(HelperStandingARF, TestARFBase):

    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.fastened_sphere = True
        self.large_viscous_boundary_layer = False
        self.small_viscous_boundary_layer = True
        self.assign_parameters()

    def compute_arf(self):
        out = 5 / 6 + 3 / 4 * self.norm_delta

        out *= pi * self.rho_f * self.x_0**3 * self.cls.field.abs_A_squared
        out *= sin(2 * self.position)

        return out


class TestARFLimit3Traveling(HelperTravelingARF, TestARFBase):

    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.fastened_sphere = True
        self.large_viscous_boundary_layer = False
        self.small_viscous_boundary_layer = True
        self.assign_parameters()

    def compute_arf(self):
        out = 3 / 2 * pi * self.rho_f * self.x_0**3
        out *= self.cls.field.abs_A_squared

        out *= self.norm_delta * (1 - self.norm_delta)

        return out


# |x| << |x_v| << 1

class TestARFLimit4Standing(HelperStandingARF, TestARFBase):

    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.fastened_sphere = True
        self.large_viscous_boundary_layer = True
        self.small_viscous_boundary_layer = False
        self.assign_parameters()

    def compute_arf(self):
        out = 1 + 10 / 27 / self.norm_delta

        out *= pi * self.rho_f * self.x_0**3 * self.cls.field.abs_A_squared
        out *= 0.9 * sin(2 * self.position) * self.norm_delta

        return out


class TestARFLimit4Traveling(HelperTravelingARF, TestARFBase):

    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.fastened_sphere = True
        self.large_viscous_boundary_layer = True
        self.small_viscous_boundary_layer = False
        self.assign_parameters()

    def compute_arf(self):
        out = -3 / 2 * pi * self.rho_f * self.x_0**3
        out *= self.cls.field.abs_A_squared

        out *= self.norm_delta**2 * (1 - 6 / 5 / self.norm_delta)

        return out


if __name__ == '__main__':
    unittest.main()



import matplotlib.pyplot as plt
import numpy as np

from osaft.plotting.base_plotting import BasePlot
from osaft.plotting.datacontainers.arf_datacontainer import ARFData


class ARFPlotter(BasePlot):

    def __init__(self) -> None:
        self._abscissa = None
        self._attr_name = None

    def plot_solution(
            self, x_values: np.ndarray, data: ARFData, file_name: str,
            ax: plt.Axes, plot_method, **kwargs,
    ) -> (plt.Figure, plt.Axes):
        fig, ax = self._create_figure(ax)

        plt.sca(ax)
        plot_method(
            x_values, data.plotting,
            label=data.name,
            linestyle=data.line_style,
            **kwargs,
        )

        return fig, ax

    @staticmethod
    def add_legend(ax: plt.Axes) -> None:
        ax.legend()

    @staticmethod
    def set_labels(
        ax: plt.Axes,
        attr_name: str,
        norm_name: str = None,
    ) -> None:
        ax.set_xlabel(f'Attribute {attr_name}')

        if norm_name is None:
            ax.set_ylabel(r'$F^{\mathrm{rad}}$ $\mathrm{[N]}$')
            ax.set_title('ARF for different solutions')
        else:
            ax.set_ylabel('normalized ARF [-]')
            ax.set_title(
                f'Normalized ARF by {norm_name} for different solutions',
            )


if __name__ == '__main__':
    pass



from osaft import WaveTypes, Yosioka1955
from osaft.tests.basetest_plotting import BaseTestPlotting


class BaseTestScattering(BaseTestPlotting):

    def setUp(self) -> None:

        super().setUp()
        self.f = 10e6
        self.R_0 = 30e-6
        self.c_s = 3e2
        self.c_f = 1.5e3
        self.rho_s = 1e3
        self.rho_f = 1.5e3
        self.p_0 = 1e5
        self.wave_type = WaveTypes.TRAVELLING
        self.position = self.c_f / self.f / 8

        self.N_max = 10

        self.yosioka = Yosioka1955.ScatteringField(
            frequency=self.f,
            radius=self.R_0,
            rho_s=self.rho_s, c_s=self.c_s,
            rho_f=self.rho_f, c_f=self.c_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
            N_max=self.N_max,
        )


if __name__ == '__main__':
    pass



from __future__ import annotations

from typing import Optional, Union

import numpy as np

from osaft import log
from osaft.core.backgroundfields import WaveTypes, WrongWaveTypeError
from osaft.core.basecomposite import BaseSphereFrequencyComposite
from osaft.core.frequency import Frequency
from osaft.core.functions import conj, pi, sin, xexpexp1
from osaft.core.geometries import Sphere
from osaft.core.helper import StringFormatter as SF
from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.solutions.base_arf import BaseARF
from osaft.solutions.Doinikov1994Rigid.scattering import ScatteringField


class ARF(ScatteringField, BaseARF, BaseSphereFrequencyComposite):
    """ARF class for Doinikov (viscous fluid-rigid sphere; 1994)

    .. note::
       For all models the condition :math:`|x| \\ll 1` must hold.

       If :attr:`small_viscous_boundary_layer` is set to `True` than in
       addition :math:`|x| \\ll |x_v| \\ll 1` must hold. If set to
       `False` than :math:`|x| \\ll 1 \\ll |x_v|` must hold.

       Both values will be logged in the `info.log` when calling
       :meth:`acoustic_radiation_force`. These conditions will not be
       checked at runtime.


    :param frequency: Frequency [Hz]
    :param radius: Radius of the sphere [m]
    :param rho_s: Density of the fluid-like sphere [kg/m^3]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of wave, travel(l)ing or standing
    :param position: Position in the standing wave field [rad]
    :param small_particle_limit: using the limiting cases for the ARF
    :param small_viscous_boundary_layer: :math:`x_v \\ll 1`
    :param large_viscous_boundary_layer: :math`x_v \\gg 1`
    :param fastened_sphere: use theory of fastened sphere
    :param N_max: Highest order mode
    """

    def __init__(
        self, frequency: Union[Frequency, float, int],
        radius: Union[Sphere, float, int],
        rho_s: float,
        rho_f: float, c_f: float,
        eta_f: float, zeta_f: float,
        p_0: float,
        wave_type: WaveTypes = WaveTypes.STANDING,
        position: Optional[float] = None,
        small_particle_limit: bool = True,
        small_viscous_boundary_layer: bool = False,
        large_viscous_boundary_layer: bool = False,
        fastened_sphere: bool = False,
        N_max: int = 5,
    ) -> None:
        """Constructor method
        """

        # init of parent class
        BaseSphereFrequencyComposite.__init__(self, frequency, radius)
        ScatteringField.__init__(
            self, frequency, radius, rho_s, rho_f,
            c_f, eta_f, zeta_f,
            p_0, wave_type, position, N_max,
        )

        # independent variables
        self._fastened_sphere = PassiveVariable(
            fastened_sphere,
            'fastened sphere limit',
        )

        self._check_small_particle_limit_value(small_particle_limit)
        self._small_particle_limit = PassiveVariable(
            small_particle_limit,
            'small particle limit',
        )
        self._small_viscous_boundary_layer = PassiveVariable(
            small_viscous_boundary_layer,
            'small viscous boundary layer limit',
        )

        self._large_viscous_boundary_layer = PassiveVariable(
            large_viscous_boundary_layer,
            'large viscous boundary layer limit',
        )

        # dependent variables
        self._norm_delta = ActiveVariable(
            self._compute_norm_delta,
            'normalized boundary layer thickness',
        )
        self._norm_delta.is_computed_by(
            self.fluid._delta, self.sphere._R_0,
        )

        # Coefficients G_n
        self._G_0 = ActiveVariable(self._compute_G_0, 'G_0')
        self._G_1 = ActiveVariable(self._compute_G_1, 'G_1')
        self._G_2 = ActiveVariable(self._compute_G_2, 'G_2')
        self._G_3 = ActiveVariable(self._compute_G_3, 'G_3')
        self._G_4 = ActiveVariable(self._compute_G_4, 'G_4')

        self._G_0.is_computed_by(self._rho_t, self._x_v)
        self._G_1.is_computed_by(self._x_v)
        self._G_2.is_computed_by(self._x_v)
        self._G_3.is_computed_by(self._x_v)
        self._G_4.is_computed_by(self._x_v)

        # Coefficients D_n
        self._D_0 = ActiveVariable(self._compute_D_0, 'D_0')
        self._D_1 = ActiveVariable(self._compute_D_1, 'D_1')

        self._D_0.is_computed_by(self._x, self._x_v, self._G_0, self._G_1)
        self._D_1.is_computed_by(
            self._x, self._x_v, self._G_0, self._G_2, self._G_3, self._G_4,
        )

        # Streaming contribution to the force F_2
        self._F_2_standing = ActiveVariable(
            self._compute_F_2_standing,
            'F_2 standing wave',
        )
        self._F_2_travelling = ActiveVariable(
            self._compute_F_2_travelling,
            'F_2 travelling wave',
        )
        self._F_2_standing.is_computed_by(
            self._x, self._x_v,
            self.fluid._rho_f,
            self.fluid._k_f,
            self.field._abs_pos,
            self.field._abs_A_squared,
        )
        self._F_2_travelling.is_computed_by(
            self._x, self._x_v,
            self.fluid._rho_f,
            self.fluid._k_f,
            self.field._abs_pos,
            self.field._abs_A_squared,
        )

        # logging
        log.debug(repr(self))
        log.info(str(self))

    def __repr__(self):
        return (
            f'Donikov1994Rigid.ARF(frequency={self.f}, radius={self.R_0}, '
            f'rho_s={self.rho_s}, rho_f={self.rho_f}, c_f={self.c_f}, '
            f'eta_f={self.eta_f}, zeta_f={self.zeta_f}, '
            f'p_0={self.p_0}, position={self.position}, {self.wave_type}, '
            f'small_viscous_boundary_layer='
            f'{self.small_viscous_boundary_layer}, '
            f'large_viscous_boundary_layer='
            f'{self.large_viscous_boundary_layer}, '
            f'small_particle_limit={self.small_particle_limit}, '
            f'fastened_sphere= {self.fastened_sphere}, '
            f'N_max={self.N_max})'
        )

    def __str__(self):
        out = 'Doinikovs\'s  (1994) model (viscous fluid-rigid sphere) for the'
        out += ' ARF with the following properties: \n'
        out += 'Limit Cases\n'
        out += SF.get_str_text(
            'Small particle limit',
            'small_particle_limit',
            self.small_particle_limit, None,
        )
        out += SF.get_str_text(
            'Small delta',
            'small_viscous_boundary_layer',
            self.small_viscous_boundary_layer, None,
        )
        out += SF.get_str_text(
            'Fastened Sphere',
            'fastened_sphere',
            self.fastened_sphere, None,
        )
        out += 'Backgroundfield\n'
        out += SF.get_str_text('Frequency', 'f', self.f, 'Hz')
        out += SF.get_str_text('Pressure', 'p_0', self.p_0, 'Pa')
        out += SF.get_str_text('Wavetype', 'wave_type', self.wave_type, None)
        out += SF.get_str_text(
            'Position', 'd',
            self.position, 'rad',
        )
        out += SF.get_str_text(
            'Wavelength', 'lambda',
            self.field.lambda_f, 'm',
        )

        out += 'Fluid\n'
        out += SF.get_str_text(
            'Density', 'rho_f',
            self.rho_f, 'kg/m^3',
        )
        out += SF.get_str_text(
            'Sound Speed', 'c_0',
            self.c_f, 'm/s',
        )
        out += SF.get_str_text(
            'Compressibility', 'kappa_f',
            self.kappa_f, '1/Pa',
        )
        out += SF.get_str_text(
            'Shear viscosity', 'eta_f',
            self.eta_f, '1/Pa',
        )
        out += SF.get_str_text(
            'Bulk viscosity', 'zeta_f',
            self.zeta_f, '1/Pa',
        )

        out += 'Particle\n'
        out += SF.get_str_text(
            'Radius', 'R_0',
            self.R_0, 'm',
        )
        out += SF.get_str_text(
            'Density', 'rho_s',
            self.rho_s, 'kg/m^3',
        )
        return out

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def acoustic_radiation_force(self) -> float:
        """Acoustic radiation fore in [N]

        It logs the current values of :attr:`x` and :attr:`x_v`.

        :raises ValueError: Non supported :attr:`~osaft.solutions.\
Doinikov1994Rigid.baseclass.aseDoinikov1994Rigid.wave_type`
        """
        if not isinstance(self.wave_type, WaveTypes):
            raise WrongWaveTypeError

        log.info(
            'Computing the ARF with Doinikov1994Rigid.\n'
            f'x   = {self.x:+.4e} -> |x|   = {np.absolute(self.x):+.4e}\n'
            f'x_v = {self.x_v:+.4e} -> |x_v| = {np.absolute(self.x_v):+.4e}',
        )

        # Small Particle Solutions
        if self.fastened_sphere:
            if self.large_viscous_boundary_layer:
                return self._arf_fastened_large_delta()
            else:
                return self._arf_fastened_small_delta()
        if self.small_viscous_boundary_layer:
            return self._arf_small_particle_small_delta()
        elif self.large_viscous_boundary_layer:
            return self._arf_small_particle_large_delta()
        else:
            return self._arf_small_particle()

    # -------------------------------------------------------------------------
    # Special Expressions for the ARF
    # -------------------------------------------------------------------------

    def _arf_fastened_small_delta(self) -> float:
        # eq. (7.1) for travelling wave and
        # eq. (6.14) w/ \\delta -> 0 for standing
        if self.wave_type == WaveTypes.TRAVELLING:
            out = 1 - self.norm_delta
            out *= 1.5 * self.norm_delta
        else:
            out = 5 / 6 + 0.75 * self.norm_delta
            out *= self._standing_wave_contribution()
        out *= pi * self.rho_f * self.field.abs_A_squared * self.x_0**3
        return out

    def _arf_fastened_large_delta(self) -> float:
        # eq. (7.5) for travelling wave and eq. (7.6) for standing
        if self.wave_type == WaveTypes.TRAVELLING:
            out = 1 - 1.2 / self.norm_delta

            out *= -1.5 * self.norm_delta
        else:
            out = 1 + 10 / 27 / self.norm_delta
            out *= 0.9 * self._standing_wave_contribution()
        out *= pi * self.rho_f * self.field.abs_A_squared * self.x_0**3
        out *= self.norm_delta
        return out

    def _arf_small_particle_large_delta(self) -> float:
        # eq. (6.18) for travelling wave and eq. (6.19) for standing
        if self.wave_type == WaveTypes.TRAVELLING:
            out = (1 - self.rho_t) / self.norm_delta
            out *= -22 / 15
        else:
            out = 11 * (1 - self.rho_t) / 5 / self.norm_delta
            out += 2 * self.rho_t - 1
            out *= self._standing_wave_contribution() / 3
        out *= pi * self.rho_s * self.field.abs_A_squared * self.x_0**3
        return out

    def _arf_small_particle_small_delta(self) -> float:
        # eq. (6.13) for travelling wave and eq. (6.14) for standing
        if self.wave_type == WaveTypes.TRAVELLING:
            out = 2 + 12 * self.rho_t - 21 * self.rho_t**2 + 7 * self.rho_t**3
            out *= -self.norm_delta / (2 + self.rho_t)
            out += (1 - self.rho_t)**2
            out *= 6 * self.norm_delta
            out /= (2 + self.rho_t)**2
        else:
            out = 3 * (1 - self.rho_t)**2 * self.norm_delta
            out /= (2 + self.rho_t)**2
            out += (5 - 2 * self.rho_t) / (3 * (2 + self.rho_t))
            out *= self._standing_wave_contribution()
        out *= pi * self.rho_f * self.field.abs_A_squared * self.x_0**3
        return out

    def _arf_small_particle(self) -> float:
        """ Acoustic radiation force for the case where the particle radius
        and the viscous boundary layer are small compared to the wavelength
        """
        if self.wave_type == WaveTypes.TRAVELLING:
            return self._arf_small_particle_travelling()
        else:
            return self._arf_small_particle_standing()

    def _arf_small_particle_travelling(self) -> float:
        """Eq. (5.10), (5.12), (6.1), (6.2)
        """
        # F1
        F1 = self.D_0 - self.D_0.conjugate()
        F1 += 2 * (self.D_1 - self.D_1.conjugate())
        F1 *= 3 / 2 * 1j * pi * self.rho_f * self.field.abs_A_squared
        # F_2
        F2 = self.F_2_travelling
        return (F1 + F2).real

    def _arf_small_particle_standing(self) -> float:
        """Eq. (5.15), (5.16), (6.1), (6.2)
        """
        # F1
        sine = sin(2 * self.k_f * self.abs_pos)
        sine_conj = sin(
            2 * self.k_f.conjugate() *
            self.abs_pos,
        )
        F1 = self.D_0 * sine + self.D_0.conjugate() * sine_conj
        F1 -= 2 * (self.D_1 * sine + self.D_1.conjugate() * sine_conj)
        F1 *= 3 / 4 * pi * self.rho_f * self.field.abs_A_squared
        # F_2
        F2 = self.F_2_standing
        return (F1 + F2).real

    def _standing_wave_contribution(self) -> float:
        return sin(2 * self.position)

    # -------------------------------------------------------------------------
    # Streaming Contribution to the ARF
    # -------------------------------------------------------------------------

    @property
    def F_2_standing(self) -> complex:
        """Returns the streaming contribution to the ARF for a standing wave

        (eq 5.17)
        """
        return self._F_2_standing.value

    def _compute_F_2_standing(self) -> complex:
        x = self.x
        x_con = self.x.conjugate()

        first = x - x_con
        first *= sin(
            self.abs_pos * (
                self.k_f +
                conj(self.k_f)
            ),
        )
        first *= sin(x + x_con) / (self.x + x_con)

        second = x + x_con
        second *= sin(
            self.abs_pos * (
                self.k_f -
                conj(self.k_f)
            ),
        )
        second *= sin(x - x_con) / (x - x_con)

        F2 = (first - second)
        F2 *= x * x_con / self.x_v**2
        F2 *= - 3 / 4 * pi * self.rho_f * self.field.abs_A_squared

        return F2

    @property
    def F_2_travelling(self) -> complex:
        """Returns the streaming contribution to the ARF for a standing wave

        (eq 5.12)
        """
        return self._F_2_travelling.value

    def _compute_F_2_travelling(self) -> complex:
        x = self.x
        x_con = self.x.conjugate()

        F2 = x * x_con * (x + x_con) * 1j / self.x_v**2
        F2 *= sin(x - x_con) / (x - x_con)
        F2 *= - 3 / 2 * pi * self.rho_f * self.field.abs_A_squared
        return F2

    # -------------------------------------------------------------------------
    # Coefficients D_n
    # -------------------------------------------------------------------------

    @property
    def D_0(self) -> complex:
        """Approximation for the coefficient D_0 from Eq. (6.1)
        """
        return self._D_0.value

    def _compute_D_0(self) -> complex:
        out = self.G_1
        out += self.x_v ** 3 * (12 + self.x_v ** 2) * self._f(self.x_v)
        out *= - self.x ** 3 * self.G_0
        out += 2 * self.x ** 3 / 9
        out += self.x ** 3 / (3 * self.x_v ** 2)
        return out

    @property
    def D_1(self) -> complex:
        """Approximation for the coefficient D_1 from Eq. (6.2)
        """
        return self._D_1.value

    def _compute_D_1(self):
        out = - 2 * (1 + 1j) * (9 + self.x_v ** 2)
        out *= self._f(self.x_v - 1j * self.x_v)
        out += 2 * self.G_3 * self._f(self.x_v)
        out += 1j * self.G_4 * self._f(-1j * self.x_v)
        out *= - self.x_v ** 3
        if abs(out.real + self.G_2.real) / abs(self.G_2.real) < 1e-6:
            RuntimeWarning(  # pragma: no cover
                'Possible precision loss detected. '
                'Consider using a limiting case solution.',
            )
        out += self.G_2
        out *= self.x ** 3 * self.G_0.conjugate() / (1 + self.x_v)
        out += self.x ** 3 / (3 * self.x_v ** 2)
        return out

    @staticmethod
    def _f(z):
        return xexpexp1(z)

    # -------------------------------------------------------------------------
    # Coefficients G_n
    # -------------------------------------------------------------------------

    @property
    def G_0(self) -> complex:
        """Approximation to value G_0 from Eq. (6.3)
        """
        return self._G_0.value

    def _compute_G_0(self) -> complex:
        denominator = 9 * self.rho_t + 9 * self.rho_t * self.x_v
        denominator += (2 + self.rho_t) * self.x_v ** 2
        out = (1 - self.rho_t) / (72 * denominator)
        return out

    @property
    def G_1(self) -> complex:
        """Approximation to value G_1 from Eq. (6.4)
        """
        return self._G_1.value

    def _compute_G_1(self) -> complex:
        out = 48 - 96 * self.x_v + 2 * self.x_v ** 2 - 14 * self.x_v ** 3
        out += self.x_v ** 4 - self.x_v ** 5
        return out

    @property
    def G_2(self) -> complex:
        """Approximation to value G_2 from Eq. (6.5)
        """
        return self._G_2.value

    def _compute_G_2(self) -> complex:
        out = 48 + (48 + 192 * 1j / 5) * self.x_v
        out += (122 + 192 * 1j) * self.x_v ** 2 / 5
        out += 42 * (1 - 1j) * self.x_v ** 3 / 5
        out += (49 + 36 * 1j) * self.x_v ** 4 / 10
        out -= (31 - 17 * 1j) * self.x_v ** 5 / 10
        out += (6 + 31 * 1j) * self.x_v ** 6 / 30
        out += (1 + 6 * 1j) * self.x_v ** 7 / 30
        out += 1j * self.x_v ** 8 / 30
        return out

    @property
    def G_3(self) -> complex:
        """Approximation to value G_3 from Eq. (6.6)
        """
        return self._G_3.value

    def _compute_G_3(self) -> complex:
        return 3 - 3 * 1j * self.x_v - self.x_v ** 2

    @property
    def G_4(self) -> complex:
        """Approximation to value G_4 from Eq. (6.7)
        """
        return self._G_4.value

    def _compute_G_4(self) -> complex:
        out = 9 + 9 * self.x_v + 41 * self.x_v ** 2 / 10
        out += 11 * self.x_v ** 3 / 10 + self.x_v ** 4 / 5 + self.x_v ** 5 / 30
        return out

    # -------------------------------------------------------------------------
    # Exceptions
    # -------------------------------------------------------------------------

    @staticmethod
    def _check_small_particle_limit_value(value):
        if not value:
            raise NotImplementedError(
                'You are trying to use the general solution to the ARF '
                'for this model. However, so far just the limiting cases '
                'are implemented. Feel free to help us to extend this.',
            )

    # -------------------------------------------------------------------------
    # Getters and Setters
    # -------------------------------------------------------------------------

    @property
    def small_particle_limit(self) -> bool:
        """Use limiting case for ARF calculation

        :getter: returns if a small particle limit is used
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        :raises NotImplementedError: If value is set to False; because just the
            limiting cases for the ARF are implemented
        """
        return self._small_particle_limit.value

    @small_particle_limit.setter
    def small_particle_limit(self, value):
        self._check_small_particle_limit_value(value)
        self._small_particle_limit.value = value

    @property
    def small_viscous_boundary_layer(self) -> bool:
        """Use limiting case of a small viscous boundary layer :math:`\\delta`

        :getter: returns if a small viscous boundary layer case is used
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._small_viscous_boundary_layer.value

    @small_viscous_boundary_layer.setter
    def small_viscous_boundary_layer(self, value):
        self._small_viscous_boundary_layer.value = value

    @property
    def large_viscous_boundary_layer(self) -> bool:
        """Use limiting case of a large viscous boundary layer :math:`\\delta`

        :getter: returns if a large viscous boundary layer case is used
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._large_viscous_boundary_layer.value

    @large_viscous_boundary_layer.setter
    def large_viscous_boundary_layer(self, value):
        self._large_viscous_boundary_layer.value = value

    @property
    def fastened_sphere(self) -> bool:
        """Use limiting case of fastened sphere

        :getter: returns if fastened sphere limiting case is used
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._fastened_sphere.value

    @fastened_sphere.setter
    def fastened_sphere(self, value):
        self._fastened_sphere.value = value

    @property
    def norm_delta(self) -> float:
        """ normalized viscous boundary thickness
        :math:`\\tilde{\\delta}=\\frac{\\delta}{R_0}`
        """
        return self._norm_delta.value

    def _compute_norm_delta(self):
        return self.delta / self.R_0

    @property
    def x_0(self) -> float:
        """ Real part of :attr:`~.x`
        """
        return self.x.real


if __name__ == '__main__':
    pass



from typing import Union

from osaft import log
from osaft.core.basecomposite import BaseFrequencyComposite
from osaft.core.frequency import Frequency
from osaft.core.functions import sqrt
from osaft.core.variable import ActiveVariable, PassiveVariable


class RigidSolid(BaseFrequencyComposite):
    """RigidSolid class

    :param frequency: excitation frequency in [Hz]
    :param rho: density in [kg/m^3]
    """

    def __init__(
        self, frequency: Union[int, float, Frequency],
        rho: float,
    ) -> None:
        """Constructor method
        """

        # Calling parent class
        super().__init__(frequency)

        # Independent variables
        self._rho_s = PassiveVariable(rho, 'density rho')

        if type(self) is RigidSolid:
            log.debug(f'Creating {self}')

    def __repr__(self) -> str:
        return f'RigidSolid(f={self.f}, rho={self.rho_s})'

    # -------------------------------------------------------------------------
    # Setters and Getters for Independent Variables
    # -------------------------------------------------------------------------

    @property
    def rho_s(self) -> float:
        """Density of the solid :math:`\\rho_{s}`.

        :getter: returns the value for the density
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._rho_s.value

    @rho_s.setter
    def rho_s(self, value: float) -> None:
        self._rho_s.value = value


class ElasticSolid(RigidSolid):
    """ElasticSolid class

    :param frequency: excitation frequency in [Hz]
    :param E: Young's modulus in [Pa]
    :param nu: Poisson's ratio in [-]
    :param rho: density in [km m^-3]
    """

    def __init__(self, frequency, E, nu, rho) -> None:
        """Constructor method
        """

        super().__init__(frequency, rho)

        # Independent variables
        self._E_s = PassiveVariable(E, "Young's modulus E")
        self._nu_s = PassiveVariable(nu, "Poisson's ratio nu")

        # Dependent variables
        self._c_l = ActiveVariable(self._compute_c_l, 'primary wave speed c_l')
        self._k_l = ActiveVariable(
            self._compute_k_l,
            'wavenumber of primary wave k_l',
        )

        self._c_t = ActiveVariable(
            self._compute_c_t,
            'secondary wave speed c_t',
        )
        self._k_t = ActiveVariable(
            self._compute_k_t,
            'wavenumber of secondary wave k_t',
        )

        self._B_s = ActiveVariable(self._compute_B_s, 'bulk modulus B_s')
        self._kappa_s = ActiveVariable(
            self._compute_kappa_s,
            'compressibility kappa_s',
        )

        # Dependencies
        self._c_l.is_computed_by(self._E_s, self._nu_s, self._rho_s)
        self._k_l.is_computed_by(self._c_l, self.frequency._omega)

        self._c_t.is_computed_by(self._E_s, self._nu_s, self._rho_s)
        self._k_t.is_computed_by(self._c_t, self.frequency._omega)

        self._B_s.is_computed_by(self._E_s, self._nu_s)
        self._kappa_s.is_computed_by(self._B_s)

        # no if clause neede because last in inheritance diagram
        log.debug(f'Creating {self}')

    def __repr__(self) -> str:
        return (
            f'ElasticSolid(f={self.f}, rho={self.rho_s}, '
            f'E={self.E_s}, nu={self.nu_s})'
        )

    # -------------------------------------------------------------------------
    # Setters and Getters for Independent Variables
    # -------------------------------------------------------------------------

    @property
    def E_s(self) -> float:
        """Young's modulus of the solid :math:`E_{s}` [Pa].

        :getter: returns the value for the Young's modulus
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._E_s.value

    @E_s.setter
    def E_s(self, value: float) -> None:
        self._E_s.value = value

    @property
    def nu_s(self) -> float:
        """Poisson's ratio of the solid :math:`\\nu_{s}`.

        :getter: returns the value for the Poisson's ratio
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._nu_s.value

    @nu_s.setter
    def nu_s(self, value: float) -> None:
        self._nu_s.value = value

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def B_s(self) -> float:
        """Returns the bulk modulus of the solid :math:`B_{s}` [Pa]
        """
        return self._B_s.value

    @property
    def kappa_s(self) -> float:
        """Returns the compressibility of the solid :math:`\\kappa_{s}` [1/Pa]
        """
        return self._kappa_s.value

    @property
    def c_l(self) -> float:
        """Returns the longitudinal wave speed in the solid :math:`c_l` [m/s]
        """
        return self._c_l.value

    @property
    def c_t(self) -> float:
        """Return the transverse wave speed in the solid :math:`c_t` [m/s]
        """
        return self._c_t.value

    @property
    def k_l(self) -> float:
        """Returns the longitudinal wave number in the solid :math:`k_l` [1/m]
        """
        return self._k_l.value

    @property
    def k_t(self) -> float:
        """Returns the transverse wave number in the solid :math:`k_t` [1/m]
        """
        return self._k_t.value

    # -------------------------------------------------------------------------
    # Dependent Variables Methods
    # -------------------------------------------------------------------------

    def _compute_B_s(self) -> float:
        return self.E_s / (1 - 2 * self.nu_s) / 3

    def _compute_kappa_s(self) -> float:
        return 1 / self.B_s

    def _compute_c_l(self) -> float:
        return sqrt(
            self.E_s * (1 - self.nu_s) / (
                self.rho_s * (1 + self.nu_s)
                * (1 - 2 * self.nu_s)
            ),
        )

    def _compute_c_t(self) -> float:
        return sqrt(self.E_s / (2 * (1 + self.nu_s) * self.rho_s))

    def _compute_k_l(self) -> float:
        return self.omega / self.c_l

    def _compute_k_t(self) -> float:
        return self.omega / self.c_t


if __name__ == '__main__':
    pass



import codecs
import os.path
import re

from setuptools import find_packages, setup

with open('requirements.txt') as f:
    install_requires = f.read().splitlines()

with open('requirements_dev.txt') as f:
    dev_requires = f.read().splitlines()

with open('docs/requirements.txt') as f:
    docs_requires = f.read().splitlines()

with open('README.md') as f:
    long_description = f.read()


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        match = re.search(r'## (\d{1,2}\.\d{1,2}\.\d{1,2})', line)
        if match is not None:
            match = match.group(1)
            return str(match)
    else:
        raise RuntimeError('Unable to find version string.')


setup(
    name='osaft',
    version=get_version('CHANGELOG.md'),
    packages=find_packages(),
    description='An Open-Source Python Framework For Acoustofluidics',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='osaftpy developer team',
    author_email='osaftpy@gmail.com',
    url='https://gitlab.com/acoustofluidics/osaft',
    project_urls={
        'Documentation': 'https://osaft.readthedocs.io/en/stable/',
        'Tracker': 'https://gitlab.com/acoustofluidics/osaft/-/issues',
    },
    install_requires=install_requires,
    extras_require={
        'dev': dev_requires,
        'docs': docs_requires,
    },
    test_suite='tests',
    keywords='msaf arf scattering acoustofluidic acoustic-radiation-force',
    license='LPGL v3',
    python_requires='>=3.9',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU Lesser General Public License v3 '
        '(LGPLv3)',
        'Programming Language :: Python :: 3.9',
    ],
)



from typing import Union

import numpy as np

from osaft.solutions.base_arf import BaseARF


class ARFData:
    """Container for plotting data for ARF plots

    :param sol: solution
    """

    _instances = 0
    _line_styles = [
        '-',
        '--',
        ':',
        '-.',
        (0, (5, 10)),
        (0, (3, 1, 1, 1)),
        (0, (1, 2)),
        (0, (8, 10)),
        (0, (5, 5)),
        (0, (5, 1)),
        (0, (3, 10, 1, 10)),
        (0, (3, 5, 1, 5)),
        (0, (3, 1, 1, 1)),
        (0, (3, 10, 1, 10, 1, 10)),
        (0, (3, 5, 1, 5, 1, 5)),
        (0, (3, 1, 1, 1, 1, 1)),
    ]

    def __init__(self, sol: BaseARF) -> None:
        self.sol = sol
        self._arf = None
        self._norm_arf = None
        self._instance = ARFData._instances
        ARFData._instances += 1

    def compute_arf(self, attr_name: str, values: np.ndarray) -> None:
        arf = []
        for val in values:
            setattr(self.sol, attr_name, val)
            arf.append(self.sol.acoustic_radiation_force())

        self._arf = np.asarray(arf)
        self._plotting = self._arf

    def normalize_arf(self, norm: np.ndarray) -> None:
        self._norm_arf = self._arf / norm
        self._plotting = self._norm_arf

    @property
    def line_style(self) -> Union[str, tuple[int, tuple[int]]]:

        L = len(ARFData._line_styles)
        return ARFData._line_styles[self._instance % L]

    @property
    def arf(self) -> np.ndarray:
        return self._arf

    @property
    def plotting(self) -> np.ndarray:
        return self._plotting

    @property
    def name(self) -> str:
        return self.sol.name


if __name__ == '__main__':
    pass



from numbers import Number
from typing import Union

from osaft import log
from osaft.core.frequency import Frequency
from osaft.core.geometries import Sphere


class BaseFrequencyComposite:
    """ Base class for all composite classes with
    :class:`~osaft.core.frequency.Frequency`

    :param frequency: excitation frequency in [Hz]

    This class has two purposes:

        #. Provides an `__init__` methods that handles the `frequency` argument
           correctly to make sure that all components and the composite have
           the same instance of :class:`~osaft.core.frequency.Frequency`.
        #. Provides wrappers for setters and getters of f and omega
    """

    def __init__(self, frequency: Union[int, float, Frequency]) -> None:
        """Constructor method
        """

        if isinstance(frequency, Number):
            log.debug(
                f'Passed frequency of type {type(frequency)}.'
                f'New Frequency object is created.',
            )
            self.frequency = Frequency(frequency)
        elif isinstance(frequency, Frequency):
            log.debug(
                f'Passed frequency of type {type(frequency)}.'
                f'No new Frequency object is created',
            )
            self.frequency = frequency
        else:
            raise TypeError(
                f'Expected type of frequency argument to be '
                f'int, float, or Frequency.'
                f'got {type(frequency)}.',
            )

        if type(self) is BaseFrequencyComposite:
            log.debug(f'Creating {self}')

    def __repr__(self) -> str:
        """ Repr string for debugging

        """
        return f'BaseFrequencyComposite(f={self.f})'

    # -------------------------------------------------------------------------
    # Attribute Wrappers
    # -------------------------------------------------------------------------

    @property
    def f(self) -> float:
        """ wrapper for :attr:`osaft.core.frequency.Frequency.f`
        """
        return self.frequency.f

    @f.setter
    def f(self, value) -> None:
        self.frequency.f = value

    @property
    def omega(self) -> float:
        """ wrapper for :attr:`osaft.core.frequency.Frequency.omega`
        """
        return self.frequency.omega


class BaseSphereFrequencyComposite(BaseFrequencyComposite):
    """ Composite class with :class:`~osaft.core.frequency.Frequency` and
    :class:`~osaft.core.geometries.Sphere`

    This class has two purposes:

        #. Provides an `__init__` methods that handles the `frequency`
           argument and the `radius` argument correctly to make sure that all
           components and the composite have the same instance of
           :class:`~osaft.core.frequency.Frequency`
           and :class:`~osaft.core.geometries.Sphere`.
        #. Provides wrappers for setters and getters.

    :param frequency: excitation frequency in [Hz]
    :param radius: sphere radius in [m]
    """

    def __init__(
        self, frequency: Union[int, float, Frequency],
        radius: Union[int, float, Sphere],
    ) -> None:
        """Constructor method
        """

        super().__init__(frequency)

        if isinstance(radius, Number):
            log.debug(
                f'Passed frequency of type {type(radius)}.'
                f'New Frequency object is created.',
            )
            self.sphere = Sphere(radius)
        elif isinstance(radius, Sphere):
            log.debug(
                f'Passed sphere of type {type(radius)}.'
                f'No new Sphere object is created',
            )
            self.sphere = radius
        else:
            raise TypeError(
                f'Expected type of radius argument to be'
                f'int, float, or Sphere.'
                f'got {type(radius)}.',
            )

        if type(self) is BaseSphereFrequencyComposite:
            log.debug(f'Creating {self}')

    def __repr__(self) -> str:
        return (
            f'BaseSphereFrequencyComposite(f={self.f}, '
            f'r={self.R_0})'
        )

    # -------------------------------------------------------------------------
    # Attribute Wrappers
    # -------------------------------------------------------------------------

    @property
    def R_0(self) -> float:
        """ Wrapper for :attr:`osaft.core.geometries.Sphere.R_0`
        """
        return self.sphere.R_0

    @R_0.setter
    def R_0(self, value) -> None:
        self.sphere.R_0 = value

    @property
    def A(self) -> float:
        """ Wrapper for :attr:`osaft.core.geometries.Sphere.A`
        """
        return self.sphere.A

    @property
    def V(self) -> float:
        """ Wrapper for :attr:`osaft.core.geometries.Sphere.V`
        """
        return self.sphere.V


if __name__ == '__main__':
    pass



from __future__ import annotations

from enum import Enum, auto
from typing import Optional

from numpy import absolute

from osaft import log
from osaft.core.functions import exp, full_range
from osaft.core.generic_types import t_fluid
from osaft.core.variable import ActiveVariable, PassiveVariable


class DocEnum(Enum):
    def __new__(cls, value, doc=None):
        self = object.__new__(cls)
        self._value_ = value
        self.__doc__ = doc
        return self


class WaveTypes(DocEnum):
    """Supported Wavetypes for the acoustic backgorundfield
    """

    STANDING = auto(), 'Acoustic standing wave'
    TRAVELLING = auto(), 'Acoustic travelling wave'


class WrongWaveTypeError(Exception):
    """ Exception for using the wrong Wave type
    """

    def __init__(self, msg=None, *args, **kwargs):
        if msg is None:
            msg = (
                'Wavetype needs to be of type:'
                f' {[t.name for t in WaveTypes]}'
            )
        super().__init__(msg, *args, **kwargs)


class BackgroundField:
    """BackgroundField class which is a base class for all background field
    classes

    :param fluid: fluid for to be used for the background field
    :param p_0: pressure amplitude in [Pa]
    :param wave_type: wave type: either standing or traveling / travelling
    :param position: position of the scatterer in the standing wave in [rad]
    """

    def __init__(
        self, fluid: t_fluid,
        p_0: float,
        wave_type: WaveTypes = WaveTypes.STANDING,
        position: Optional[float] = None,
    ) -> None:
        """Constructor Method
        """

        self.fluid = fluid

        # Check position variable
        if position is None and wave_type == WaveTypes.STANDING:
            raise ValueError(
                'Argument position needs to be given for a '
                'standing wave.',
            )

        # Independent Variables
        self._p_0 = PassiveVariable(p_0, 'pressure amplitude p_0')
        self._wave_type = PassiveVariable(wave_type, 'wave type')
        self._position = PassiveVariable(position, 'scatterer position d')

        # Dependent variables
        self._abs_pos = ActiveVariable(
            self._compute_abs_pos,
            'absolute position in the standing wave',
        )

        self._A = ActiveVariable(
            self._compute_A,
            'velocity potential amplitude A',
        )
        self._abs_A_squared = ActiveVariable(
            self._compute_abs_A_squared,
            'squared absolute amplitude A',
        )
        self._A_in = ActiveVariable(
            self._reset_A_in,
            'velocity potential amplitude A_n',
        )
        self._E = ActiveVariable(
            self._compute_E,
            'acoustic energy density E',
        )
        self._I = ActiveVariable(
            self._compute_I,
            'acoustic intensity I',
        )
        # Dependencies
        self._abs_pos.is_computed_by(self.fluid._k_f, self._position)

        self._A.is_computed_by(
            self.fluid.frequency._omega,
            self.fluid._rho_f,
            self._p_0,
        )

        self._abs_A_squared.is_computed_by(self._A)

        self._A_in.is_computed_by(
            self._abs_pos,
            self._A,
            self._wave_type,
        )

        self._E.is_computed_by(
            self._A,
            self.fluid._rho_f,
            self.fluid._k_f,
        )

        self._I.is_computed_by(
            self._A,
            self.fluid._rho_f,
            self.fluid._k_f,
            self.fluid._c_f,
        )

        log.debug(f'Creating {self}')

    def __repr__(self) -> str:
        return (
            f'BackgroundField(fluid={self.fluid}, p_0={self.p_0} '
            f'wave_type={self.wave_type}, position={self.position})'
        )

    # -------------------------------------------------------------------------
    # Setters and Getters for Independent Variables
    # -------------------------------------------------------------------------

    @property
    def p_0(self) -> float:
        """Background pressure amplitude of the acoustic field [Pa]

        :getter: returns the value for the pressure amplitude
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._p_0.value

    @p_0.setter
    def p_0(self, value: float) -> None:
        self._p_0.value = value

    @property
    def wave_type(self) -> WaveTypes:
        """Wave type of the acoustic background field

        :getter: returns the wave type
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._wave_type.value

    @wave_type.setter
    def wave_type(self, value: WaveTypes) -> None:
        self._wave_type.value = value

    @property
    def position(self) -> float:
        """Position of the particle in an acoustic standing wave field [rad]

        :getter: returns the wave type
        :setter: automatically invokes
           :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._position.value

    @position.setter
    def position(self, value: float) -> None:
        self._position.value = value

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def abs_pos(self) -> float:
        """Absolute position of the scatterer in the standing wave in [m] """
        return self._abs_pos.value

    @property
    def A(self) -> complex:
        """Amplitude of the pressure wave [Pa]
        """
        return self._A.value

    @property
    def abs_A_squared(self) -> float:
        """:math:`A\\,\\text{conj}(A) = |A|^2` [Pa^2]
        """
        return self._abs_A_squared.value

    @property
    def E_ac(self) -> float:
        """Mean acoustic energy density [J/m^3]
        """
        if not isinstance(self.wave_type, WaveTypes):
            raise WrongWaveTypeError

        return self._E.value

    @property
    def I_ac(self) -> float:
        """Acoustic intensity / energy flux density [W/m^2]
        """
        if self.wave_type != WaveTypes.TRAVELLING:
            raise WrongWaveTypeError(
                'The acoustic intensity is only '
                'defined for: '
                f'{WaveTypes.TRAVELLING}',
            )
        return self._I.value

    # -------------------------------------------------------------------------
    # Method for Dependent Variables
    # -------------------------------------------------------------------------

    def _compute_A(self) -> complex:
        return self.p_0 / (1j * self.omega * self.rho_f)

    def _compute_abs_A_squared(self) -> complex:
        return absolute(self.A)**2

    @staticmethod
    def _reset_A_in() -> list:
        return []

    # -------------------------------------------------------------------------
    # Wrapper Methods for Frequency
    # -------------------------------------------------------------------------

    @property
    def f(self) -> float:
        """ wrapper for :attr:`osaft.core.fluids.InviscidFluid.f`
        """
        return self.fluid.f

    @f.setter
    def f(self, value) -> None:
        self.fluid.f = value

    @property
    def omega(self) -> float:
        """ wrapper for :attr:`osaft.core.fluids.InviscidFluid.omega`
        """
        return self.fluid.omega

    # -------------------------------------------------------------------------
    # Wrapper Methods for Independent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def c_f(self) -> float:
        """ wrapper for :attr:`osaft.core.fluids.InviscidFluid.c_f`
        """
        return self.fluid.c_f

    @c_f.setter
    def c_f(self, value) -> None:
        self.fluid.c_f = value

    @property
    def rho_f(self) -> float:
        """ wrapper for :attr:`osaft.core.fluids.InviscidFluid.rho_f`
        """
        return self.fluid.rho_f

    @rho_f.setter
    def rho_f(self, value) -> None:
        self.fluid.rho_f = value

    @property
    def eta_f(self) -> float:
        """ wrapper for :attr:`osaft.core.fluids.ViscousFluid.eta_f`
        """
        try:
            return self.fluid.eta_f
        except AttributeError:
            log.warning(
                'underlying fluid class of BackgroundField doesn\'t '
                'have attribute eta_f',
            )
            return 0

    @eta_f.setter
    def eta_f(self, value) -> None:
        if hasattr(self.fluid, 'eta_f'):
            self.fluid.eta_f = value
        else:
            log.warning(
                'underlying fluid class of BackgroundField doesn\'t '
                'have attribute eta_f',
            )

    @property
    def zeta_f(self) -> float:
        """ wrapper for :attr:`osaft.core.fluids.ViscousFluid.zeta_f`
        """
        try:
            return self.fluid.zeta_f
        except AttributeError:
            log.warning(
                'underlying fluid class of BackgroundField doesn\'t '
                'have attribute zeta_f',
            )
            return 0

    @zeta_f.setter
    def zeta_f(self, value) -> None:
        if hasattr(self.fluid, 'zeta_f'):
            self.fluid.zeta_f = value
        else:
            log.warning(
                'underlying fluid class of BackgroundField doesn\'t '
                'have attribute zeta_f',
            )

    @property
    def eta_p(self) -> float:
        """ wrapper for :attr:`osaft.core.fluids.ViscoelasticFluid.eta_p`
        """
        try:
            return self.fluid.eta_p
        except AttributeError:
            log.warning(
                'underlying fluid class of BackgroundField doesn\'t '
                'have attribute eta_p',
            )
            return 0

    @eta_p.setter
    def eta_p(self, value) -> None:
        if hasattr(self.fluid, 'eta_p'):
            self.fluid.eta_p = value
        else:
            log.warning(
                'underlying fluid class of BackgroundField doesn\'t '
                'have attribute eta_p',
            )

    @property
    def zeta_p(self) -> float:
        """ wrapper for :attr:`osaft.core.fluids.ViscoelasticFluid.zeta_p`
        """
        try:
            return self.fluid.zeta_p
        except AttributeError:
            log.warning(
                'underlying fluid class of BackgroundField doesn\'t '
                'have attribute zeta_p',
            )
            return 0

    @zeta_p.setter
    def zeta_p(self, value) -> None:
        if hasattr(self.fluid, 'zeta_p'):
            self.fluid.zeta_p = value
        else:
            log.warning(
                'underlying fluid class of BackgroundField doesn\'t '
                'have attribute zeta_p',
            )

    @property
    def lambda_M(self) -> float:
        """ wrapper for :attr:`osaft.core.fluids.ViscoelasticFluid.lambda_M`
        """
        try:
            return self.fluid.lambda_M
        except AttributeError:
            log.warning(
                'underlying fluid class of BackgroundField doesn\'t '
                'have attribute lambda_M',
            )
            return 0

    @lambda_M.setter
    def lambda_M(self, value) -> None:
        if hasattr(self.fluid, 'lambda_M'):
            self.fluid.lambda_M = value
        else:
            log.warning(
                'underlying fluid class of BackgroundField doesn\'t '
                'have attribute lambda_M',
            )

    # -------------------------------------------------------------------------
    # Wrapper Methods for Dependent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def lambda_f(self) -> float:
        """ wrapper for :attr:`osaft.core.fluids.InviscidFluid.lambda_f`
        """
        return self.fluid.lambda_f

    @property
    def k_f(self) -> float:
        """ wrapper for :attr:`osaft.core.fluids.InviscidFluid.k_f`
        """
        return self.fluid.k_f

    @property
    def k_v(self) -> complex:
        """ wrapper for :attr:`osaft.core.fluids.ViscousFluid.k_v`
        """
        return self.fluid.k_v

    @property
    def kappa_f(self) -> float:
        """ wrapper for :attr:`osaft.core.fluids.ViscousFluid.kappa_f`
        """
        return self.fluid.kappa_f

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def _compute_abs_pos(self) -> float:
        return self.position / self.k_f.real

    def _compute_A_in(self, N: int) -> None:
        if not isinstance(self.wave_type, WaveTypes):
            raise WrongWaveTypeError

        N_old = len(self._A_in.value)

        if self.wave_type == WaveTypes.TRAVELLING:
            self.__compute_A_in_travelling(N, N_old)
        else:
            self.__compute_A_in_standing(N, N_old)

    def __compute_A_in_travelling(self, N: int, N_old: int) -> None:
        for n in full_range(N_old, N):
            self._A_in.value.append(self.A * (2 * n + 1) * 1j ** n)

    def __compute_A_in_standing(self, N: int, N_old: int) -> None:
        for n in full_range(N_old, N):
            new = exp(1j * self.abs_pos * self.k_f)
            new += (-1) ** n * exp(-1j * self.abs_pos * self.k_f)

            new *= self.A / 2 * (2 * n + 1) * 1j ** n
            self._A_in.value.append(new)

    def A_in(self, n: int) -> complex:
        """Incident amplitude at order n [Pa]

        :param n: order
        """
        if n < len(self._A_in.value):
            return self._A_in.value[n]
        else:
            self._compute_A_in(n)
            return self._A_in.value[n]

    def _compute_E(self) -> float:
        return (self.fluid.kappa_f * self.p_0**2) / 4

    def _compute_I(self) -> float:
        out = self.abs_A_squared * self.rho_f * self.fluid.c_f
        out *= (self.fluid.k_f ** 2) / 2
        return out


if __name__ == '__main__':
    pass



import numpy as np

from osaft.core.functions import SpecialFunctions as Sp
from osaft.core.functions import full_range
from osaft.core.variable import ActiveVariable


class CoefficientMatrix:
    """Matrix containing for the computation of scattering coefficients
    """

    def __init__(self) -> None:
        self._m_11 = ActiveVariable(self._reset_list, 'm_11 entry of matrix')
        self._m_12 = ActiveVariable(self._reset_list, 'm_12 entry of matrix')
        self._m_13 = ActiveVariable(self._reset_list, 'm_13 entry of matrix')
        self._m_14 = ActiveVariable(self._reset_list, 'm_14 entry of matrix')

        self._m_21 = ActiveVariable(self._reset_list, 'm_21 entry of matrix')
        self._m_22 = ActiveVariable(self._reset_list, 'm_22 entry of matrix')
        self._m_23 = ActiveVariable(self._reset_list, 'm_23 entry of matrix')
        self._m_24 = ActiveVariable(self._reset_list, 'm_24 entry of matrix')

        self._m_31 = ActiveVariable(self._reset_list, 'm_31 entry of matrix')
        self._m_32 = ActiveVariable(self._reset_list, 'm_32 entry of matrix')
        self._m_33 = ActiveVariable(self._reset_list, 'm_33 entry of matrix')
        self._m_34 = ActiveVariable(self._reset_list, 'm_34 entry of matrix')

        self._m_41 = ActiveVariable(self._reset_list, 'm_41 entry of matrix')
        self._m_42 = ActiveVariable(self._reset_list, 'm_42 entry of matrix')
        self._m_43 = ActiveVariable(self._reset_list, 'm_43 entry of matrix')
        self._m_44 = ActiveVariable(self._reset_list, 'm_44 entry of matrix')

        self._n_1 = ActiveVariable(self._reset_list, 'n_1 entry of RHS')
        self._n_2 = ActiveVariable(self._reset_list, 'n_2 entry of RHS')
        self._n_3 = ActiveVariable(self._reset_list, 'n_3 entry of RHS')
        self._n_4 = ActiveVariable(self._reset_list, 'n_4 entry of RHS')

    def det_M(self, n: int) -> complex:
        """det(M) of order :attr:`n`

        :param n: order
        """
        return np.linalg.det(self.M(n))

    def det_M_1(self, n: int) -> complex:
        """det(M_1) of order :attr:`n`

        :param n: order
        """
        return np.linalg.det(self.M_col(n, 1))

    def det_M_2(self, n: int) -> complex:
        """det(M_2) of order :attr:`n`

        :param n: order
        """
        return np.linalg.det(self.M_col(n, 2))

    def det_M_3(self, n: int) -> complex:
        """det(M_3) of order :attr:`n`

        :param n: order
        """
        return np.linalg.det(self.M_col(n, 3))

    def det_M_4(self, n: int) -> complex:
        """det(M_4) of order :attr:`n`

        :param n: order
        """
        return np.linalg.det(self.M_col(n, 4))

    def M(self, n: int) -> np.ndarray:
        """Matrix M of order :attr:`n`

        :param n: order
        """

        out = np.ndarray((4, 4), dtype=np.complex_)

        for i in full_range(1, 4):
            for j in full_range(1, 4):
                out[i - 1, j - 1] = getattr(self, f'm_{i}{j}')(n)

        return out

    def M_col(self, n: int, col: int) -> np.ndarray:
        """Matrix :math:`M_{col}` of order :attr:`n`

        :param n: order
        :param col: column that is replaced by RHS vector N
        """
        out = np.ndarray((4, 4), dtype=np.complex_)

        for i in full_range(1, 4):
            for j in full_range(1, 4):
                if j != col:
                    out[i - 1, j - 1] = getattr(self, f'm_{i}{j}')(n)
                else:
                    out[i - 1, j - 1] = getattr(self, f'n_{i}')(n)
        return out

    # -----------------------------------------------------------------------
    # RHS entries
    # -----------------------------------------------------------------------

    def n_1(self, n: int) -> complex:
        """Coefficient n_1 of RHS vector N of :attr:`n` th order

        :param n: order
        """
        if n < len(self._n_1.value):
            return self._n_1.value[n]
        else:
            self._compute_n_1(n)
            return self._n_1.value[n]

    def _compute_n_1(self, N: int) -> None:

        N_old = len(self._n_1.value)
        for n in full_range(N_old, N):
            out = -self.x * Sp.d1_besselj(n, self.x)
            out *= self.A_in(n)
            self._n_1.value.append(out)

    def n_2(self, n: int) -> complex:
        """Coefficient n_2 of RHS vector N of :attr:`n` th order

        :param n: order
        """
        if n < len(self._n_2.value):
            return self._n_2.value[n]
        else:
            self._compute_n_2(n)
            return self._n_2.value[n]

    def _compute_n_2(self, N: int) -> None:

        N_old = len(self._n_2.value)
        for n in full_range(N_old, N):
            out = -Sp.besselj(n, self.x)
            out *= self.A_in(n)
            self._n_2.value.append(out)

    def n_3(self, n: int) -> complex:
        """Coefficient n_3 of RHS vector N of :attr:`n` th order

        :param n: order
        """
        if n < len(self._n_3.value):
            return self._n_3.value[n]
        else:
            self._compute_n_3(n)
            return self._n_3.value[n]

    def _compute_n_3(self, N: int) -> None:

        N_old = len(self._n_3.value)
        for n in full_range(N_old, N):
            out = 1j * self.rho_f * self.c_f**2 / self.omega
            out += self.zeta_f
            out -= 2 / 3 * self.eta_f
            out *= Sp.besselj(n, self.x)

            out -= 2 * self.eta_f * Sp.d2_besselj(n, self.x)

            out *= self.A_in(n) * self.x**2
            self._n_3.value.append(out)

    def n_4(self, n: int) -> None:
        """Coefficient n_4 of RHS vector N of :attr:`n` th order

        :param n: order
        """
        if n < len(self._n_4.value):
            return self._n_4.value[n]
        else:
            self._compute_n_4(n)
            return self._n_4.value[n]

    def _compute_n_4(self, N: int) -> None:

        N_old = len(self._n_4.value)
        for n in full_range(N_old, N):
            out = Sp.besselj(n, self.x)
            out -= self.x * Sp.d1_besselj(n, self.x)

            out *= 2 * self.eta_f * self.A_in(n)
            self._n_4.value.append(out)
    # -----------------------------------------------------------------------
    # Matrix entries m_1j
    # -----------------------------------------------------------------------

    def m_11(self, n: int) -> None:
        """Coefficient m_11 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_11.value):
            return self._m_11.value[n]
        else:
            self._compute_m_11(n)
            return self._m_11.value[n]

    def _compute_m_11(self, N: int) -> None:
        N_old = len(self._m_11.value)
        for n in full_range(N_old, N):
            out = self.x * Sp.d1_hankelh1(n, self.x)

            self._m_11.value.append(out)

    def m_12(self, n: int) -> None:
        """Coefficient m_12 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_12.value):
            return self._m_12.value[n]
        else:
            self._compute_m_12(n)
            return self._m_12.value[n]

    def _compute_m_12(self, N: int) -> None:
        N_old = len(self._m_12.value)
        for n in full_range(N_old, N):
            out = -n * (n + 1) * Sp.hankelh1(n, self.x_v)

            self._m_12.value.append(out)

    def m_13(self, n: int) -> None:
        """Coefficient m_13 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_13.value):
            return self._m_13.value[n]
        else:
            self._compute_m_13(n)
            return self._m_13.value[n]

    def _compute_m_13(self, N: int) -> None:
        N_old = len(self._m_13.value)
        for n in full_range(N_old, N):
            out = - self.x_hat * Sp.d1_besselj(n, self.x_hat)

            self._m_13.value.append(out)

    def m_14(self, n: int) -> None:
        """Coefficient m_14 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_14.value):
            return self._m_14.value[n]
        else:
            self._compute_m_14(n)
            return self._m_14.value[n]

    def _compute_m_14(self, N: int) -> None:
        N_old = len(self._m_14.value)
        for n in full_range(N_old, N):
            out = n * (n + 1) * Sp.besselj(n, self.x_hat_v)

            self._m_14.value.append(out)

    # -----------------------------------------------------------------------
    # Matrix entries m_2j
    # -----------------------------------------------------------------------

    def m_21(self, n: int) -> None:
        """Coefficient m_21 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_21.value):
            return self._m_21.value[n]
        else:
            self._compute_m_21(n)
            return self._m_21.value[n]

    def _compute_m_21(self, N: int) -> None:
        N_old = len(self._m_21.value)
        for n in full_range(N_old, N):
            out = Sp.hankelh1(n, self.x)

            self._m_21.value.append(out)

    def m_22(self, n: int) -> None:
        """Coefficient m_22 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_22.value):
            return self._m_22.value[n]
        else:
            self._compute_m_22(n)
            return self._m_22.value[n]

    def _compute_m_22(self, N: int) -> None:
        N_old = len(self._m_22.value)
        for n in full_range(N_old, N):
            out = -Sp.hankelh1(n, self.x_v)
            out -= self.x_v * Sp.d1_hankelh1(n, self.x_v)

            self._m_22.value.append(out)

    def m_23(self, n: int) -> None:
        """Coefficient m_23 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_23.value):
            return self._m_23.value[n]
        else:
            self._compute_m_23(n)
            return self._m_23.value[n]

    def _compute_m_23(self, N: int) -> None:
        N_old = len(self._m_23.value)
        for n in full_range(N_old, N):
            out = -Sp.besselj(n, self.x_hat)

            self._m_23.value.append(out)

    def m_24(self, n: int) -> None:
        """Coefficient m_24 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_24.value):
            return self._m_24.value[n]
        else:
            self._compute_m_24(n)
            return self._m_24.value[n]

    def _compute_m_24(self, N: int) -> None:
        N_old = len(self._m_24.value)
        for n in full_range(N_old, N):
            out = Sp.besselj(n, self.x_hat_v)
            out += self.x_hat_v * Sp.d1_besselj(n, self.x_hat_v)

            self._m_24.value.append(out)

    # -----------------------------------------------------------------------
    # Matrix entries m_3j
    # -----------------------------------------------------------------------

    def m_31(self, n: int) -> None:
        """Coefficient m_31 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_31.value):
            return self._m_31.value[n]
        else:
            self._compute_m_31(n)
            return self._m_31.value[n]

    def _compute_m_31(self, N: int) -> None:
        N_old = len(self._m_31.value)
        for n in full_range(N_old, N):
            out = 1j * self.rho_f * self.c_f**2 / self.omega

            out += self.zeta_f
            out -= 2 / 3 * self.eta_f
            out *= -Sp.hankelh1(n, self.x)

            out += 2 * self.eta_f * Sp.d2_hankelh1(n, self.x)

            out *= self.x**2

            self._m_31.value.append(out)

    def m_32(self, n: int) -> None:
        """Coefficient m_32 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_32.value):
            return self._m_32.value[n]
        else:
            self._compute_m_32(n)
            return self._m_32.value[n]

    def _compute_m_32(self, N: int) -> None:
        N_old = len(self._m_32.value)
        for n in full_range(N_old, N):
            out = Sp.hankelh1(n, self.x_v)
            out -= self.x_v * Sp.d1_hankelh1(n, self.x_v)

            out *= 2 * n * (n + 1) * self.eta_f

            self._m_32.value.append(out)

    def m_33(self, n: int) -> None:
        """Coefficient m_33 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_33.value):
            return self._m_33.value[n]
        else:
            self._compute_m_33(n)
            return self._m_33.value[n]

    def _compute_m_33(self, N: int) -> None:
        N_old = len(self._m_33.value)
        for n in full_range(N_old, N):
            out = 1j * self.rho_s * self.c_s**2 / self.omega
            out += self.zeta_s
            out -= 2 / 3 * self.eta_s
            out *= Sp.besselj(n, self.x_hat)

            out -= 2 * self.eta_s * Sp.d2_besselj(n, self.x_hat)

            out *= self.x_hat**2

            self._m_33.value.append(out)

    def m_34(self, n: int) -> None:
        """Coefficient m_34 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_34.value):
            return self._m_34.value[n]
        else:
            self._compute_m_34(n)
            return self._m_34.value[n]

    def _compute_m_34(self, N: int) -> None:
        N_old = len(self._m_34.value)
        for n in full_range(N_old, N):
            out = self.x_hat_v * Sp.d1_besselj(n, self.x_hat_v)
            out -= Sp.besselj(n, self.x_hat_v)

            out *= 2 * n * (n + 1) * self.eta_s

            self._m_34.value.append(out)

    # -----------------------------------------------------------------------
    # Matrix entries m_4j
    # -----------------------------------------------------------------------

    def m_41(self, n: int) -> None:
        """Coefficient m_41 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_41.value):
            return self._m_41.value[n]
        else:
            self._compute_m_41(n)
            return self._m_41.value[n]

    def _compute_m_41(self, N: int) -> None:
        N_old = len(self._m_41.value)
        for n in full_range(N_old, N):
            out = self.x * Sp.d1_hankelh1(n, self.x)
            out -= Sp.hankelh1(n, self.x)

            out *= 2 * self.eta_f

            self._m_41.value.append(out)

    def m_42(self, n: int) -> None:
        """Coefficient m_42 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_42.value):
            return self._m_42.value[n]
        else:
            self._compute_m_42(n)
            return self._m_42.value[n]

    def _compute_m_42(self, N: int) -> None:
        N_old = len(self._m_42.value)
        for n in full_range(N_old, N):
            out = self.x_v**2 * Sp.d2_hankelh1(n, self.x_v)
            out += (n**2 + n - 2) * Sp.hankelh1(n, self.x_v)

            out *= -self.eta_f

            self._m_42.value.append(out)

    def m_43(self, n: int) -> None:
        """Coefficient m_43 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_43.value):
            return self._m_43.value[n]
        else:
            self._compute_m_43(n)
            return self._m_43.value[n]

    def _compute_m_43(self, N: int) -> None:
        N_old = len(self._m_43.value)
        for n in full_range(N_old, N):
            out = Sp.besselj(n, self.x_hat)
            out -= self.x_hat * Sp.d1_besselj(n, self.x_hat)

            out *= 2 * self.eta_s

            self._m_43.value.append(out)

    def m_44(self, n: int) -> None:
        """Coefficient m_44 of M matrix of :attr:`n` th order

        :param n: order
        """
        if n < len(self._m_44.value):
            return self._m_44.value[n]
        else:
            self._compute_m_44(n)
            return self._m_44.value[n]

    def _compute_m_44(self, N: int) -> None:
        N_old = len(self._m_44.value)
        for n in full_range(N_old, N):
            out = self.x_hat_v**2 * Sp.d2_besselj(n, self.x_hat_v)
            out += (n**2 + n - 2) * Sp.besselj(n, self.x_hat_v)

            out *= self.eta_s

            self._m_44.value.append(out)



import unittest

from osaft import Doinikov1994Compressible, Settnes2012, WaveTypes
from osaft.tests.basetest import BaseTest
from osaft.tests.basetest_arf import HelperCompareARF


@unittest.skip('')
class TestCompareSmallBoundaryLayer(BaseTest, HelperCompareARF):

    def setUp(self):
        super().setUp()

        self.arf_compare_threshold = 1e-2
        self.small_viscous_boundary_layer = True
        self.large_viscous_boundary_layer = False
        self.small_particle_limit = True
        self.wave_type = WaveTypes.TRAVELLING
        self._wave_type.list_of_values = [WaveTypes.TRAVELLING]

        # Frequency
        self._f.low = 1e6
        self._f.high = 3e6
        self.f = 1e6

        # Radius
        self._R_0.high = 5e-6
        self._R_0.low = 3e-6
        self.R_0 = 3e-6

        # Viscosity
        self._eta_f.low = 1e-4
        self._eta_f.high = 1e-2
        self.eta_f = 1e-4
        self.zeta_f = 0
        self.eta_s = 1
        self.zeta_s = 1

        self.cls = Doinikov1994Compressible.ARF(
            self.f,
            self.R_0,
            self.rho_s, self.c_s, self.eta_s, self.zeta_s,
            self.rho_f, self.c_f, self.eta_f, self.zeta_f,
            self.p_0, self.wave_type,
            self.position,
            small_particle_limit=self.small_viscous_boundary_layer,
            large_viscous_boundary_layer=self.large_viscous_boundary_layer,
            small_viscous_boundary_layer=self.small_particle_limit,
        )

        print(self.cls.x_v)
        print(self.cls.x)

        self.compare_cls = Settnes2012.ARF(
            self.f,
            self.R_0,
            self.rho_s, self.c_s,
            self.rho_f, self.c_f, self.eta_f,
            self.p_0, self.wave_type,
            self.position,
        )

        self.list_cls = [self.cls, self.compare_cls]


if __name__ == '__main__':
    pass



import unittest

from osaft.core.basecomposite import (
    BaseFrequencyComposite,
    BaseSphereFrequencyComposite,
)
from osaft.core.frequency import Frequency
from osaft.core.functions import pi
from osaft.core.geometries import Sphere
from osaft.tests.basetest import BaseTest


class TestBaseFrequencyComposite(BaseTest):

    def test_types_for_frequency(self):
        cls = BaseFrequencyComposite(10.1)
        self.assertTrue(isinstance(cls, BaseFrequencyComposite))
        cls = BaseFrequencyComposite(10.1)
        self.assertTrue(isinstance(cls, BaseFrequencyComposite))

        F = Frequency(1e6)
        cls = BaseFrequencyComposite(F)
        self.assertTrue(isinstance(cls, BaseFrequencyComposite))

    def test_TypeError(self):

        self.assertRaises(TypeError, BaseFrequencyComposite, 's')


class TestBaseSphereFrequencyComposite(BaseTest):

    def test_types_for_sphere(self):
        cls = BaseSphereFrequencyComposite(10.1, 10)
        self.assertTrue(isinstance(cls, BaseSphereFrequencyComposite))
        cls = BaseSphereFrequencyComposite(10.1, 10.0)
        self.assertTrue(isinstance(cls, BaseSphereFrequencyComposite))

        S = Sphere(1e-6)
        cls = BaseSphereFrequencyComposite(10.1, S)
        self.assertTrue(isinstance(cls, BaseSphereFrequencyComposite))

    def test_properties(self):
        def V(R):
            return 4 / 3 * R**3 * pi

        def A(R):
            return 4 * R**2 * pi

        R = 1e-6
        f = 1e6
        cls = BaseSphereFrequencyComposite(f, R)

        self.assertAlmostEqual(R, cls.R_0)
        self.assertAlmostEqual(V(R), cls.V)
        self.assertAlmostEqual(A(R), cls.A)

        R = 4.39e-6
        cls.R_0 = R
        self.assertAlmostEqual(R, cls.R_0)
        self.assertAlmostEqual(V(R), cls.V)
        self.assertAlmostEqual(A(R), cls.A)

    def test_TypeError(self):
        self.assertRaises(
            TypeError, BaseSphereFrequencyComposite,
            10.0, 's',
        )


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from typing import Optional, Union

import numpy as np

from osaft.core.functions import exp, pi
from osaft.core.functions import spherical_2_cartesian_coordinates as s2c_coord
from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.plotting.datacontainers.base_velocity_datacontainer import (
    VelocityFieldData,
)
from osaft.solutions.base_scattering import BaseScattering


class ScatteringFieldData(VelocityFieldData, ABC):
    """Container for plotting data for the scattering field

    :param sol: solution
    :param r_min: start point for radial plotting range
    :param r_max: end point for tangential plotting range
    :param theta_min: start point for tangential plotting range
    :param theta_max: end point for tangential plotting range
    :param res: resolution, if tuple (radial resolution, tangential resolution)
    """

    def __init__(
        self,
        sol: BaseScattering,
        r_min: float,
        r_max: float,
        theta_min: float = 0,
        theta_max: float = pi,
        res: Union[int, tuple[int, int]] = 100,
    ) -> None:
        """Constructor method"""

        VelocityFieldData.__init__(
            self, sol, r_min, r_max, theta_min,
            theta_max, res,
        )

        # Velocities for the scattering field (complex valued)
        self._u = ActiveVariable(self._reset_complex_array, 'x coordinate')
        self._w = ActiveVariable(self._reset_complex_array, 'z coordinate')

        self._u.is_computed_by(self._sol, self._grid, self._mode)
        self._w.is_computed_by(self._sol, self._grid, self._mode)
    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def get_velocity_magnitude(
            self,
            instantaneous: bool = True,
            phase: float = 0,
            mode: Optional[int] = None,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Returns the velocity magnitudes

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the magnitude of the velocity field at
        these points

        :param instantaneous: if `True` the instant magnitude is returned
        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        """
        self._mode.value = mode
        if instantaneous:
            return self.x, self.z, self._velocity_norm_inst(phase)
        else:
            return self.x, self.z, self.u_norm

    def get_displacement_magnitude(
            self,
            instantaneous: bool = True,
            phase: float = 0,
            mode: Optional[int] = None,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Returns the displacement magnitudes

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the magnitude of the displacement field at
        these points

        :param instantaneous: if `True` the instant magnitude is returned
        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        """
        self._mode.value = mode
        if instantaneous:
            return self.x, self.z, self._disp_norm_inst(phase)
        else:
            return self.x, self.z, self.u_norm / self.sol.omega

    def get_velocity_vector(
            self,
            phase: float = 0,
            mode: Optional[int] = None,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Returns the velocity vectors

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the velocity field vectors at
        these points

        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        """
        self._mode.value = mode
        if phase == 0:
            return self.x, self.z, self.u.real, self.w.real
        else:
            u = self.u * np.exp(- 1j * phase)
            w = self.w * np.exp(- 1j * phase)
            return self.x, self.w, u.real, w.real

    def get_displacement_vector(
            self,
            phase: float = 0,
            mode: Optional[int] = None,
    ) -> tuple[
        np.ndarray, np.ndarray,
        np.ndarray, np.ndarray,
    ]:
        """Returns the displacement vectors

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the displacement field vectors at
        these points

        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        """
        self._mode.value = mode
        if phase == 0:
            u = 1j * self.u / self.sol.omega
            w = 1j * self.w / self.sol.omega
        else:
            u = self.u * np.exp(- 1j * phase) / self.sol.omega
            w = self.w * np.exp(- 1j * phase) / self.sol.omega
        return self.x, self.w, u.real, w.real

    # -------------------------------------------------------------------------
    # Private Methods
    # -------------------------------------------------------------------------

    def _velocity_norm_inst(
            self,
            phase: float,
    ) -> np.ndarray:
        """Instantaneous velocity magnitude

        :param phase: phase angle
        """
        if phase == 0:
            return np.hypot(self.u.real, self.w.real)
        else:
            u = self.u * exp(-1j * phase)
            w = self.w * exp(-1j * phase)
            return np.hypot(u.real, w.real)

    def _disp_norm_inst(
        self,
        phase: float,
    ) -> np.ndarray:
        """Instantaneous displacement magnitude

        :param phase: phase angle
        """
        u = 1j * self.u * exp(-1j * phase) / self.sol.omega
        w = 1j * self.w * exp(-1j * phase) / self.sol.omega
        norm = np.hypot(u.real, w.real)
        return norm

    def _compute_data(self) -> None:
        """Compute the velocity data and store them in respective attributes

        Scattering velocity field is computed on the polar grid and transformed
        to Cartesian coordinates for plotting.
        """
        arr_r = self.grid.arr_r
        arr_theta = self.grid.arr_theta

        for index, theta in enumerate(arr_theta):

            # Indexing
            min_index = index * self.resolution
            max_index = min_index + self.resolution

            # Coordinates
            xs, zs = s2c_coord(self.grid.arr_r, theta)
            self._x.value[min_index:max_index] = xs
            self._z.value[min_index:max_index] = zs

            # Displacements Velocities
            u, w = self._compute_velocities(arr_r, theta)
            self._u.value[min_index:max_index] = u
            self._w.value[min_index:max_index] = w
            self._u_norm.value[min_index:max_index] = np.hypot(
                np.abs(u),
                np.abs(w),
            )

    @abstractmethod
    def _compute_velocities(
        self,
        arr_r: np.ndarray,
        theta: float,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Compute the velocity data and return it

        :param arr_r: array of r-coordinates
        :param theta: polar angle
        """
        pass  # pragma: no cover

# -----------------------------------------------------------------------------
# ParticleScatteringData
# -----------------------------------------------------------------------------


class ParticleScatteringData(ScatteringFieldData):
    """Container for plotting data for the displacement field in the particle

    :param sol: solution
    :param theta_min: start point for tangential plotting range
    :param theta_max: end point for tangential plotting range
    :param res: resolution, if tuple (radial resolution, tangential resolution)
    """

    def __init__(
        self,
        sol: BaseScattering,
        theta_min: float = 0,
        theta_max: float = pi,
        res: Union[int, tuple[int, int]] = 100,
    ) -> None:
        """Constructor method"""

        r_min = 1e-30
        r_max = sol.R_0
        super().__init__(
            sol, r_min, r_max, theta_min, theta_max, res,
        )

    # -------------------------------------------------------------------------
    # Private Methods
    # -------------------------------------------------------------------------

    def _compute_velocities(
        self,
        arr_r: np.ndarray,
        theta: float,
    ):
        """Compute the velocity data and return it

        :param arr_r: array of r-coordinates
        :param theta: polar angle
        """
        u = self.sol.radial_particle_velocity(arr_r, theta, 0, self.mode)
        w = self.sol.tangential_particle_velocity(arr_r, theta, 0, self.mode)
        return u, w

# -----------------------------------------------------------------------------
# FluidScatteringData
# -----------------------------------------------------------------------------


class FluidScatteringData(ScatteringFieldData):
    """Container for plotting data for the scattering field in the fluid

    :param sol: solution
    :param r_max: end point for tangential plotting range
    :param theta_min: start point for tangential plotting range
    :param theta_max: end point for tangential plotting range
    :param res: resolution, if tuple (radial resolution, tangential resolution)
    """
    def __init__(
        self,
        sol: BaseScattering,
        r_max: float,
        theta_min: float = 0,
        theta_max: float = pi,
        scattering: bool = True,
        incident: bool = True,
        res: Union[int, tuple[int, int]] = 100,
    ) -> None:
        """Constructor method"""

        r_min = sol.R_0
        super().__init__(
            sol, r_min, r_max, theta_min, theta_max, res,
        )

        self._scattered = PassiveVariable(
            scattering, 'option: plot '
            'scattering field',
        )
        self._incident = PassiveVariable(
            incident, 'option: plot '
            'incident field',
        )

        # Set links
        self._u.is_computed_by(self._scattered, self._incident)
        self._w.is_computed_by(self._scattered, self._incident)
    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def get_velocity_magnitude(
            self,
            instantaneous: bool = True,
            phase: float = 0,
            mode: Optional[int] = None,
            scattered: bool = True,
            incident: bool = True,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Returns the velocity magnitudes

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the magnitude of the velocity field at
        these points

        :param instantaneous: option if instantaneous or average magnitude
        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        :param scattered: option if scattering field is plotted
        :param incident: option if incident field is plotted
        """
        self.scattered = scattered
        self.incident = incident
        return super().get_velocity_magnitude(instantaneous, phase, mode)

    def get_velocity_vector(
            self,
            phase: float = 0,
            mode: Optional[int] = None,
            scattered: bool = True,
            incident: bool = True,
    ) -> tuple[
        np.ndarray, np.ndarray,
        np.ndarray, np.ndarray,
    ]:
        """Returns the velocity vectors

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the velocity field vectors at
        these points

        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        :param scattered: option if scattering field is plotted
        :param incident: option if incident field is plotted
        """
        self.scattered = scattered
        self.incident = incident
        return super().get_velocity_vector(phase, mode)

    def get_displacement_vector(
            self,
            phase: float = 0,
            mode: Optional[int] = None,
            scattered: bool = True,
            incident: bool = True,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Returns the displacement vectors

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the displacement field vectors at
        these points

        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        :param scattered: option if scattering field is plotted
        :param incident: option if incident field is plotted
        """
        self.scattered = scattered
        self.incident = incident
        return super().get_displacement_vector(phase, mode)

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def scattered(self) -> bool:
        """Option if scattering field is plotted

        :getter: returns True (plot field) or False (do not plot field)
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._scattered.value

    @scattered.setter
    def scattered(self, value: bool) -> None:
        self._scattered.value = value

    @property
    def incident(self) -> bool:
        """Option if incident field is plotted

        :getter: returns True (plot field) or False (do not plot field)
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._incident.value

    @incident.setter
    def incident(self, value: bool) -> None:
        self._incident.value = value

    # -------------------------------------------------------------------------
    # Private Methods
    # -------------------------------------------------------------------------

    def _compute_velocities(
            self,
            arr_r: np.ndarray,
            theta: float,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Compute the velocity data and return it

        :param arr_r: array of r-coordinates
        :param theta: polar angle
        """
        u = self.sol.radial_acoustic_fluid_velocity(
            arr_r, theta, 0,
            self.scattered,
            self.incident,
            self.mode,
        )
        w = self.sol.tangential_acoustic_fluid_velocity(
            arr_r, theta, 0,
            self.scattered,
            self.incident,
            self.mode,
        )
        return u, w


if __name__ == '__main__':
    pass
